"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.


DEP-- I need to update this string
Function that generates the symbolic code to perform the interpolation. Calling this function creates
 both the aesara functions for the potential field and the block.

Returns:
    aesara function for the potential field
    aesara function for the block
"""
import aesara
import aesara.tensor  as T
import numpy as np
import sys
from .aesara_graph import aesaraGeometry, aesaraOptions

aesara.config.openmp_elemwise_minsize = 10000
aesara.config.openmp = True

aesara.config.optimizer = 'fast_compile'
aesara.config.floatX = 'float64'
aesara.config.on_opt_error = 'ignore'

aesara.config.exception_verbosity = 'high'
aesara.config.compute_test_value = 'off'
aesara.config.profile_memory = False
aesara.config.scan.debug = False
aesara.config.profile = False


class aesaraExport(aesaraGeometry):

    def __init__(self, weights_op = None, grid: np.ndarray=None, weights=None):
        """

        Args:
            grid: if grid is passed it becomes constant
            weights:
        """

        super().__init__()

        # self.grid_val_T = aesara.shared(np.cast[dtype](np.zeros((2, 200))), 'Coordinates of the grid '
        #                                                                     'points to interpolate')

        self.grid_val_T = T.matrix('Coordinates of the grid points to interpolate')
        self.fault_matrix = T.matrix('Full block matrix for x_to_interpolate')
        # TODO: This is either shared or aesaraOP
       # if weights_op is None:
       #     self.weights = aesara.shared(np.ones(10000), 'Weights to compute')
       # else:
       #     self.weights = weights_op(self.input_parameters_weights())
       #  if weights_op is None:
       #      self.weights = T.vector('kriging weights')
       #  else:
       #      self.weights_op = weights_op
       #      self.weights = weights_op(*self.input_parameters_kriging())

        #self.fault_matrix = T.zeros((0, self.grid_val_T.shape[0] + 2 * self.len_points))

       # self.fault_drift = T.matrix('Block matrix at interface points')
      #  self.fault_mask = T.zeros((0, self.grid_val_T.shape[0] + 2 * self.len_points))

       # self.n_surface = aesara.shared(np.arange(1, 5, dtype='int32'), "ID of the surface")
      #  self.n_surface_op = self.n_surface
        #self.npf = aesara.shared(np.zeros(3, dtype='int32'), 'Number of points per surface accumulative')

        self.dot_version = False

    def set_weights(self, weights_op = None):
        if weights_op is None:
            self.weights = T.vector('kriging weights')
        else:
            self.weights_op = weights_op
            self.weights = weights_op(*self.input_parameters_kriging())


    def input_parameters_kriging(self):
        """
        Create a list with the symbolic variables to use when we compile the aesara function

        Returns:
            list: [self.dips_position_all, self.dip_angles_all, self.azimuth_all, self.polarity_all,
                   self.ref_layer_points_all, self.rest_layer_points_all]
        """
        ipl = [self.dips_position_all, self.dip_angles_all, self.azimuth_all, self.polarity_all, self.surface_points_all,
               self.fault_drift, self.number_of_points_per_surface_T_op]
        # self.ref_layer_points_all, self.rest_layer_points_all]
        return ipl

    def input_parameters_export(self):
        """
        Create a list with the symbolic variables to use when we compile the aesara function

        Returns:
            list:
        """
        ipl = [self.dips_position_all, self.surface_points_all,
               self.fault_drift, self.fault_matrix, self.weights, self.grid_val_T,
               self.number_of_points_per_surface_T_op]
        # self.ref_layer_points_all, self.rest_layer_points_all]
        return ipl

    def input_parameters_export_kriging(self):

        # ipl = [self.dips_position_all, self.dip_angles_all, self.azimuth_all, self.polarity_all,
        #        self.surface_points_all,
        #        self.fault_drift, self.number_of_points_per_surface_T_op]
        #
        # ipl = [self.dips_position_all, self.surface_points_all,
        #        self.fault_drift, self.fault_matrix, self.grid_val_T,
        #        self.number_of_points_per_surface_T_op]

        ipl = [self.dips_position_all, self.dip_angles_all, self.azimuth_all, self.polarity_all,
               self.surface_points_all,
               self.fault_drift,  self.fault_matrix, self.grid_val_T, self.number_of_points_per_surface_T_op]

        # self.ref_layer_points_all, self.rest_layer_points_all]
        return ipl

    def x_to_interpolate(self, verbose=0):
        """
        here I add to the grid points also the references points(to check the value of the potential field at the
        surface_points). Also here I will check what parts of the grid have been already computed in a previous series
        to avoid to recompute.

        Returns:
            aesara.tensor.matrix: The 3D points of the given grid plus the reference and rest points
        """

        grid_val = T.concatenate([self.grid_val_T, self.rest_layer_points_all,
                                  self.ref_layer_points_all])

        if verbose > 1:
            aesara.printing.pydotprint(grid_val, outfile="graphs/" + sys._getframe().f_code.co_name + ".png",
                                       var_with_name_simple=True)

        if 'grid_val' in self.verbose:
            grid_val = aesara.printing.Print('Points to interpolate')(grid_val)

        return grid_val

    def extend_dual_kriging(self):
        # TODO Think what object is worth to save to speed up computation
        """
        Tile the dual kriging vector to cover all the points to interpolate.So far I just make a matrix with the
        dimensions len(DK)x(grid) but in the future maybe I have to try to loop all this part so consume less memory

        Returns:
            aesara.tensor.matrix: Matrix with the Dk parameters repeated for all the points to interpolate
        """

        grid_val = self.x_to_interpolate()
        # if self.weights.get_value() is None:
        #     DK_parameters = self.solve_kriging()
        # else:
        #     DK_parameters = self.weights
        DK_parameters = self.weights
        # Creation of a matrix of dimensions equal to the grid with the weights for every point (big 4D matrix in
        # ravel form)
        # TODO IMP: Change the tile by a simple dot op -> The DOT version in gpu is slower
        DK_weights = T.tile(DK_parameters, (grid_val.shape[0], 1)).T

        if self.dot_version:
            DK_weights = DK_parameters

        return DK_weights


class aesaraExportGeo(aesaraExport):

    # def __init__(self):
    #
    #     super().__init__()

    def contribution_gradient_interface(self, grid_val=None, weights=None):
        """
        Computation of the contribution of the foliations at every point to interpolate

        Returns:
            aesara.tensor.vector: Contribution of all foliations (input) at every point to interpolate
        """
        if weights is None:
            weights = self.extend_dual_kriging()
        if grid_val is None:
            grid_val = self.x_to_interpolate()

        length_of_CG = self.matrices_shapes()[0]

        # Cartesian distances between the point to simulate and the dips
        hu_SimPoint = T.vertical_stack(
            (self.dips_position[:, 0] - grid_val[:, 0].reshape((grid_val[:, 0].shape[0], 1))).T,
            (self.dips_position[:, 1] - grid_val[:, 1].reshape((grid_val[:, 1].shape[0], 1))).T,
            (self.dips_position[:, 2] - grid_val[:, 2].reshape((grid_val[:, 2].shape[0], 1))).T
        )

        # Euclidian distances
        sed_dips_SimPoint = self.squared_euclidean_distances(self.dips_position_tiled, grid_val)
        # Gradient contribution
        sigma_0_grad = T.sum(
            (weights[:length_of_CG] *
             self.gi_reescale *
             (-hu_SimPoint *
              (sed_dips_SimPoint < self.a_T) *  # first derivative
              (- self.c_o_T * ((-14 / self.a_T ** 2) + 105 / 4 * sed_dips_SimPoint / self.a_T ** 3 -
                               35 / 2 * sed_dips_SimPoint ** 3 / self.a_T ** 5 +
                               21 / 4 * sed_dips_SimPoint ** 5 / self.a_T ** 7)))),
            axis=0)

        if self.dot_version:
            sigma_0_grad = T.dot(
                weights[:length_of_CG],
                self.gi_reescale *
                (-hu_SimPoint *
                 (sed_dips_SimPoint < self.a_T) *  # first derivative
                 (- self.c_o_T * ((-14 / self.a_T ** 2) + 105 / 4 * sed_dips_SimPoint / self.a_T ** 3 -
                                  35 / 2 * sed_dips_SimPoint ** 3 / self.a_T ** 5 +
                                  21 / 4 * sed_dips_SimPoint ** 5 / self.a_T ** 7))))

        # Add name to the aesara node
        sigma_0_grad.name = 'Contribution of the foliations to the potential field at every point of the grid'

        if str(sys._getframe().f_code.co_name) in self.verbose:
            sigma_0_grad = aesara.printing.Print('interface_gradient_contribution')(sigma_0_grad)

        return sigma_0_grad

    def contribution_interface(self, grid_val=None, weights=None):
        """
          Computation of the contribution of the surface_points at every point to interpolate

          Returns:
              aesara.tensor.vector: Contribution of all surface_points (input) at every point to interpolate
          """

        if weights is None:
            weights = self.extend_dual_kriging()
        if grid_val is None:
            grid_val = self.x_to_interpolate()
        length_of_CG, length_of_CGI = self.matrices_shapes()[:2]

        # Euclidian distances
        sed_rest_SimPoint = self.squared_euclidean_distances(self.rest_layer_points, grid_val)
        sed_ref_SimPoint = self.squared_euclidean_distances(self.ref_layer_points, grid_val)

        # Interface contribution
        sigma_0_interf = (T.sum(
            -weights[length_of_CG:length_of_CG + length_of_CGI, :] *
            (self.c_o_T * self.i_reescale * (
                    (sed_rest_SimPoint < self.a_T) *  # SimPoint - Rest Covariances Matrix
                    (1 - 7 * (sed_rest_SimPoint / self.a_T) ** 2 +
                     35 / 4 * (sed_rest_SimPoint / self.a_T) ** 3 -
                     7 / 2 * (sed_rest_SimPoint / self.a_T) ** 5 +
                     3 / 4 * (sed_rest_SimPoint / self.a_T) ** 7) -
                    ((sed_ref_SimPoint < self.a_T) *  # SimPoint- Ref
                     (1 - 7 * (sed_ref_SimPoint / self.a_T) ** 2 +
                      35 / 4 * (sed_ref_SimPoint / self.a_T) ** 3 -
                      7 / 2 * (sed_ref_SimPoint / self.a_T) ** 5 +
                      3 / 4 * (sed_ref_SimPoint / self.a_T) ** 7)))), axis=0))

        if self.dot_version:
            sigma_0_interf = (
                T.dot(-weights[length_of_CG:length_of_CG + length_of_CGI],
                      (self.c_o_T * self.i_reescale * (
                              (sed_rest_SimPoint < self.a_T) *  # SimPoint - Rest Covariances Matrix
                              (1 - 7 * (sed_rest_SimPoint / self.a_T) ** 2 +
                               35 / 4 * (sed_rest_SimPoint / self.a_T) ** 3 -
                               7 / 2 * (sed_rest_SimPoint / self.a_T) ** 5 +
                               3 / 4 * (sed_rest_SimPoint / self.a_T) ** 7) -
                              ((sed_ref_SimPoint < self.a_T) *  # SimPoint- Ref
                               (1 - 7 * (sed_ref_SimPoint / self.a_T) ** 2 +
                                35 / 4 * (sed_ref_SimPoint / self.a_T) ** 3 -
                                7 / 2 * (sed_ref_SimPoint / self.a_T) ** 5 +
                                3 / 4 * (sed_ref_SimPoint / self.a_T) ** 7))))))

        # Add name to the aesara node
        sigma_0_interf.name = 'Contribution of the surface_points to the potential field at every point of the grid'

        return sigma_0_interf

    def contribution_universal_drift(self, grid_val=None, weights=None, a=0, b=100000000):
        """
        Computation of the contribution of the universal drift at every point to interpolate

        Returns:
            aesara.tensor.vector: Contribution of the universal drift (input) at every point to interpolate
        """
        if weights is None:
            weights = self.extend_dual_kriging()
        if grid_val is None:
            grid_val = self.x_to_interpolate()

        length_of_CG, length_of_CGI, length_of_U_I, length_of_faults, length_of_C = self.matrices_shapes()

        universal_grid_surface_points_matrix = T.horizontal_stack(
            grid_val,
            (grid_val ** 2),
            T.stack((grid_val[:, 0] * grid_val[:, 1],
                     grid_val[:, 0] * grid_val[:, 2],
                     grid_val[:, 1] * grid_val[:, 2]), axis=1)).T

        # These are the magic terms to get the same as geomodeller
        i_rescale_aux = T.repeat(self.gi_reescale, 9)
        i_rescale_aux = T.set_subtensor(i_rescale_aux[:3], 1)
        _aux_magic_term = T.tile(i_rescale_aux[:self.n_universal_eq_T_op], (grid_val.shape[0], 1)).T

        # Drif contribution
        f_0 = (T.sum(
            weights[
            length_of_CG + length_of_CGI:length_of_CG + length_of_CGI + length_of_U_I] * self.gi_reescale * _aux_magic_term *
            universal_grid_surface_points_matrix[:self.n_universal_eq_T_op]
            , axis=0))

        if self.dot_version:
            f_0 = T.dot(
                weights[length_of_CG + length_of_CGI:length_of_CG + length_of_CGI + length_of_U_I],
                self.gi_reescale * _aux_magic_term *
                universal_grid_surface_points_matrix[:self.n_universal_eq_T_op])

        if not type(f_0) == int:
            f_0.name = 'Contribution of the universal drift to the potential field at every point of the grid'

        if str(sys._getframe().f_code.co_name) in self.verbose:
            f_0 = aesara.printing.Print('Universal terms contribution')(f_0)

        return f_0

    def faults_contribution(self, weights=None, a=0, b=100000000):
        """
        Computation of the contribution of the df drift at every point to interpolate. To get these we need to
        compute a whole block model with the df data

        Returns:
            aesara.tensor.vector: Contribution of the df drift (input) at every point to interpolate
        """
        if weights is None:
            weights = self.extend_dual_kriging()
        length_of_CG, length_of_CGI, length_of_U_I, length_of_faults, length_of_C = self.matrices_shapes()

        fault_matrix_selection_non_zero = self.fault_matrix[a:b] #* self.fault_mask[a:b] + 1

        f_1 = T.sum(
            weights[length_of_CG + length_of_CGI + length_of_U_I:, :] * fault_matrix_selection_non_zero, axis=0)

        if self.dot_version:
            f_1 = T.dot(
                weights[length_of_CG + length_of_CGI + length_of_U_I:], fault_matrix_selection_non_zero)

        # Add name to the aesara node
        f_1.name = 'Faults contribution'

        if str(sys._getframe().f_code.co_name) in self.verbose:
            f_1 = aesara.printing.Print('Faults contribution')(f_1)

        return f_1

    def scalar_field_loop(self, a, b, Z_x, grid_val, weights):

        sigma_0_grad = self.contribution_gradient_interface(grid_val[a:b], weights[:, a:b])
        sigma_0_interf = self.contribution_interface(grid_val[a:b], weights[:, a:b])
        f_0 = self.contribution_universal_drift(grid_val[a:b], weights[:, a:b], a, b)
        f_1 = self.faults_contribution(weights[:, a:b], a, b)

        # Add an arbitrary number at the potential field to get unique values for each of them
        partial_Z_x = (sigma_0_grad + sigma_0_interf + f_0 + f_1)
        Z_x = T.set_subtensor(Z_x[a:b], partial_Z_x)

        return Z_x

    def scalar_field_at_all(self, weights=None):
        """
        Compute the potential field at all the interpolation points, i.e. grid plus rest plus ref
        Returns:
            aesara.tensor.vector: Potential fields at all points

        """
        grid_val = self.x_to_interpolate()

        if weights is None:
            weights = self.extend_dual_kriging()

        grid_shape = T.stack([grid_val.shape[0]], axis=0)
        Z_x_init = T.zeros(grid_shape, dtype='float32')
        if 'grid_shape' in self.verbose:
            grid_shape = aesara.printing.Print('grid_shape')(grid_shape)

        steps = 1e13 / self.matrices_shapes()[-1] / grid_shape
        slices = T.concatenate((T.arange(0, grid_shape[0], steps[0], dtype='int64'), grid_shape))

        if 'slices' in self.verbose:
            slices = aesara.printing.Print('slices')(slices)

        Z_x_loop, updates3 = aesara.scan(
            fn=self.scalar_field_loop,
            outputs_info=[Z_x_init],
            sequences=[dict(input=slices, taps=[0, 1])],
            non_sequences=[grid_val, weights],
            profile=False,
            name='Looping grid',
            return_list=True)

        self.Z_x = Z_x_loop[-1][-1]
        self.Z_x.name = 'Value of the potential field at every point'

        if str(sys._getframe().f_code.co_name) in self.verbose:
            self.Z_x = aesara.printing.Print('Potential field at all points')(self.Z_x)

        return self.Z_x
    #
    # def get_scalar_field_at_surface_points_values(self):
    #
    #     # self.Z_x = self.scalar_field_at_all()
    #     self.scalar_field_at_surface_points_values = self.Z_x[-2*self.len_points: -self.len_points][self.npf]
    #     return self.scalar_field_at_surface_points_values
        # Z_x_grid = Z_x[:-2*self.len_points]
        # Z_x_sp = Z_x[-2*self.len_points:]
        # scalar_field_at_surface_points_values = Z_x_sp[: -self.len_points][self.npf]

       # return [Z_x_grid, Z_x_sp, scalar_field_at_surface_points_values]


class aesaraBlock(aesaraExport):
    def __init__(self, Z_x_op = None):

        if type(Z_x_op) == aesara.compile.builders.OpFromGraph:
            super(aesaraBlock, self).__init__()
            self.Z_x_op = Z_x_op
            self.Z_x = Z_x_op(*self.input_parameters_export_kriging())
            self.Z_x.name = 'Scalar op'
        else:
            self.Z_x = T.vector('Scalar')
         #   self.grid_val_T = T.matrix('Coordinates of the grid points to interpolate')
         #   self.weights = T.vector('kriging weights')

        # elif Z_x is None:
        #     self.Z_x = self.scalar_field_at_all

        self.n_surface = aesara.shared(np.arange(1, 5000, dtype='int32'), "ID of the surface")
        self.n_surface_op = self.n_surface

        self.npf = aesara.shared(np.zeros(3, dtype='int32'), 'Number of points per surface accumulative')
      #  self.npf_op = T.vector('Number of points per surface accumulative', dtype='int32')#self.npf[[0, -2]]
      #  self.Z_x = T.vector('Value of the potential field at every x_to_intep point')
        self.scalar_field_at_surface_points_values = self.Z_x[-2*self.len_points: -self.len_points][self.npf_op]

       # self.scalar_field_at_surface_points_values = T.vector('')
        #self.len_points = T.vector('Length of rest or ref surface points arrays')
        self.is_finite = aesara.shared(np.zeros(3, dtype='int32'), 'The series (fault) is finite')
        self.inf_factor = self.is_finite * 10
        self.is_fault = aesara.shared(np.zeros(3, dtype='int32'), 'The series is fault')

        # TODO shared or variable
        self.values_properties_op = T.matrix('Values that the blocks are taking')

    def input_parameters_block_formations(self):
        ipl = [self.values_properties_op, self.Z_x,
               self.number_of_points_per_surface_T_op,
               self.surface_points_all]

        # self.ref_layer_points_all, self.rest_layer_points_all]
        return ipl

    def input_parameters_block_faults(self):
        ipl = [self.values_properties_op, self.Z_x,
               self.number_of_points_per_surface_T_op,
               self.surface_points_all, self.grid_val_T]

        # self.ref_layer_points_all, self.rest_layer_points_all]
        return ipl

    def select_finite_faults(self):
        fault_points = T.vertical_stack(T.stack([self.ref_layer_points[0]], axis=0), self.rest_layer_points).T
        ctr = T.mean(fault_points, axis=1)
        x = fault_points - ctr.reshape((-1, 1))
        M = T.dot(x, x.T)
        U = T.nlinalg.svd(M)[2]
        rotated_x = T.dot(self.x_to_interpolate(), U)
        rotated_fault_points = T.dot(fault_points.T, U)
        rotated_ctr = T.mean(rotated_fault_points, axis=0)
        a_radius = (rotated_fault_points[:, 0].max() - rotated_fault_points[:, 0].min()) / 2 + self.inf_factor[
            self.n_surface_op[0] - 1]
        b_radius = (rotated_fault_points[:, 1].max() - rotated_fault_points[:, 1].min()) / 2 + self.inf_factor[
            self.n_surface_op[0] - 1]
        sel = T.lt((rotated_x[:, 0] - rotated_ctr[0]) ** 2 / a_radius ** 2 + (
                    rotated_x[:, 1] - rotated_ctr[1]) ** 2 / b_radius ** 2,
                   1)

        if "select_finite_faults" in self.verbose:
            sel = aesara.printing.Print("scalar_field_iter")(sel)

        return sel

    def compare(self, a, b, slice_init, Z_x, l, n_surface, drift):
        """
        Treshold of the points to interpolate given 2 potential field values. TODO: This function is the one we
        need to change for a sigmoid function

        Args:
            a (scalar): Upper limit of the potential field
            b (scalar): Lower limit of the potential field
            n_surface (scalar): Value given to the segmentation, i.e. lithology number
            Zx (vector): Potential field values at all the interpolated points

        Returns:
            aesara.tensor.vector: segmented values
        """

        slice_init = slice_init
        n_surface_0 = n_surface[:, slice_init:slice_init + 1]
        n_surface_1 = n_surface[:, slice_init + 1:slice_init + 2]
        drift = drift[:, slice_init:slice_init + 1]

        if 'compare' in self.verbose:
            a = aesara.printing.Print("a")(a)
            b = aesara.printing.Print("b")(b)
            # l = 200/ (a - b)
            slice_init = aesara.printing.Print("slice_init")(slice_init)
            n_surface_0 = aesara.printing.Print("n_surface_0")(n_surface_0)
            n_surface_1 = aesara.printing.Print("n_surface_1")(n_surface_1)
            drift = aesara.printing.Print("drift[slice_init:slice_init+1][0]")(drift)

        # drift = T.switch(slice_init == 0, n_surface_1, n_surface_0)
        #    drift = T.set_subtensor(n_surface[0], n_surface[1])

        # The 5 rules the slope of the function
        sigm = (-n_surface_0.reshape((-1, 1)) / (1 + T.exp(-l * (Z_x - a)))) - \
               (n_surface_1.reshape((-1, 1)) / (1 + T.exp(l * (Z_x - b)))) + drift.reshape((-1, 1))
        if 'sigma' in self.verbose:
            sigm = aesara.printing.Print("middle point")(sigm)
        #      n_surface = aesara.printing.Print("n_surface")(n_surface)
        return sigm

    def export_fault_block(self, Z_x = None, slope=50, offset_slope=5000):
        """
        Compute the part of the block model of a given series (dictated by the bool array yet to be computed)

        Returns:
            aesara.tensor.vector: Value of lithology at every interpolated point
        """

        if Z_x is None:
            Z_x = self.Z_x

        # Max and min values of the potential field.
        # max_pot = T.max(Z_x) + 1
        # min_pot = T.min(Z_x) - 1
        # max_pot += max_pot * 0.1
        # min_pot -= min_pot * 0.1

        # Value of the potential field at the surface_points of the computed series
        # TODO timeit
        max_pot = T.max(Z_x)
        # max_pot = aesara.printing.Print("max_pot")(max_pot)

        min_pot = T.min(Z_x)
        #     min_pot = aesara.printing.Print("min_pot")(min_pot)

        # max_pot_sigm = 2 * max_pot - self.scalar_field_at_surface_points_values[0]
        # min_pot_sigm = 2 * min_pot - self.scalar_field_at_surface_points_values[-1]

        boundary_pad = (max_pot - min_pot) * 0.01
        #l = slope / (max_pot - min_pot)  # (max_pot - min_pot)

        # This is the different line with respect layers
        l = T.switch(self.select_finite_faults(), offset_slope / (max_pot - min_pot), slope / (max_pot - min_pot))
        #  l = aesara.printing.Print("l")(l)

        # A tensor with the values to segment
        scalar_field_iter = T.concatenate((
            T.stack([max_pot + boundary_pad], axis=0),
            self.scalar_field_at_surface_points_values,
            T.stack([min_pot - boundary_pad], axis=0)
        ))

        if "scalar_field_iter" in self.verbose:
            scalar_field_iter = aesara.printing.Print("scalar_field_iter")(scalar_field_iter)

        # Here we just take the first element of values properties because at least so far we do not find a reason
        # to populate fault blocks with anything else

        n_surface_op_float_sigmoid = T.repeat(self.values_properties_op[[0], :], 2, axis=1)

        # TODO: instead -1 at the border look for the average distance of the input!
        # TODO I think should be -> n_surface_op_float_sigmoid[:, 2] - n_surface_op_float_sigmoid[:, 1]
        n_surface_op_float_sigmoid = T.set_subtensor(n_surface_op_float_sigmoid[:, 1], -1)
        # - T.sqrt(T.square(n_surface_op_float_sigmoid[0] - n_surface_op_float_sigmoid[2])))

        n_surface_op_float_sigmoid = T.set_subtensor(n_surface_op_float_sigmoid[:, -1], -1)
        # - T.sqrt(T.square(n_surface_op_float_sigmoid[3] - n_surface_op_float_sigmoid[-1])))

        drift = T.set_subtensor(n_surface_op_float_sigmoid[:, 0], n_surface_op_float_sigmoid[:, 1])

        if 'n_surface_op_float_sigmoid' in self.verbose:
            n_surface_op_float_sigmoid = aesara.printing.Print("n_surface_op_float_sigmoid") \
                (n_surface_op_float_sigmoid)

        fault_block, updates2 = aesara.scan(
            fn=self.compare,
            outputs_info=None,
            sequences=[dict(input=scalar_field_iter, taps=[0, 1]),
                       T.arange(0, n_surface_op_float_sigmoid.shape[1], 2, dtype='int64')],
            non_sequences=[Z_x, l, n_surface_op_float_sigmoid, drift],
            name='Looping compare',
            profile=False,
            return_list=False)

        # For every surface we get a vector so we need to sum compress them to one dimension
        fault_block = fault_block.sum(axis=0)

        # Add name to the aesara node
        fault_block.name = 'The chunk of block model of a specific series'
        if str(sys._getframe().f_code.co_name) in self.verbose:
            fault_block = aesara.printing.Print(fault_block.name)(fault_block)

        return fault_block

    def export_formation_block(self, Z_x = None, slope=5000,weights=None):
        """
        Compute the part of the block model of a given series (dictated by the bool array yet to be computed)

        Returns:
            aesara.tensor.vector: Value of lithology at every interpolated point
        """
        # TODO: IMP set soft max in the borders

        if Z_x is None:
            Z_x = self.Z_x

        max_pot = T.max(Z_x)
        # max_pot = aesara.printing.Print("max_pot")(max_pot)

        min_pot = T.min(Z_x)
        #     min_pot = aesara.printing.Print("min_pot")(min_pot)

        max_pot_sigm = 2 * max_pot - self.scalar_field_at_surface_points_values[0]
        min_pot_sigm = 2 * min_pot - self.scalar_field_at_surface_points_values[-1]

        boundary_pad = (max_pot - min_pot) * 0.01
        l = slope / (max_pot - min_pot)

        # A tensor with the values to segment
        scalar_field_iter = T.concatenate((
            T.stack([max_pot + boundary_pad], axis=0),
            self.scalar_field_at_surface_points_values,
            T.stack([min_pot - boundary_pad], axis=0)
        ))

        if "scalar_field_iter" in self.verbose:
            scalar_field_iter = aesara.printing.Print("scalar_field_iter")(scalar_field_iter)

        # Loop to segment the distinct lithologies

        n_surface_op_float_sigmoid = T.repeat(self.values_properties_op, 2, axis=1)

        # TODO: instead -1 at the border look for the average distance of the input!
        n_surface_op_float_sigmoid = T.set_subtensor(n_surface_op_float_sigmoid[:, 0], -1)
        # - T.sqrt(T.square(n_surface_op_float_sigmoid[0] - n_surface_op_float_sigmoid[2])))

        n_surface_op_float_sigmoid = T.set_subtensor(n_surface_op_float_sigmoid[:, -1], -1)
        # - T.sqrt(T.square(n_surface_op_float_sigmoid[3] - n_surface_op_float_sigmoid[-1])))

        drift = T.set_subtensor(n_surface_op_float_sigmoid[:, 0], n_surface_op_float_sigmoid[:, 1])

        if 'n_surface_op_float_sigmoid' in self.verbose:
            n_surface_op_float_sigmoid = aesara.printing.Print("n_surface_op_float_sigmoid") \
                (n_surface_op_float_sigmoid)

        formations_block, updates2 = aesara.scan(
            fn=self.compare,
            outputs_info=None,
            sequences=[dict(input=scalar_field_iter, taps=[0, 1]), T.arange(0, n_surface_op_float_sigmoid.shape[1],
                                                                            2, dtype='int64')],
            non_sequences=[self.Z_x, l, n_surface_op_float_sigmoid, drift],
            name='Looping compare',
            profile=False,
            return_list=False)

        # For every surface we get a vector so we need to sum compress them to one dimension
        formations_block = formations_block.sum(axis=0)

        # Add name to the aesara node
        formations_block.name = 'The chunk of block model of a specific series'
        if str(sys._getframe().f_code.co_name) in self.verbose:
            formations_block = aesara.printing.Print(formations_block.name)(formations_block)

        return formations_block


class aesaraLoop(aesaraExportGeo):
    def __init__(self, weights_op, Z_x_op, export_formation_op, export_fault_op):
        super(aesaraLoop, self).__init__()

        self.weights_op = weights_op
        self.Z_x_op = Z_x_op
        self.export_formation_op = export_formation_op
        self.export_fault_op = export_fault_op

        # FORMATIONS
        # ----------
        self.n_surface = aesara.shared(np.arange(2, 5, dtype='int32'), "ID of the surface")
        self.n_surface_op = self.n_surface
        self.surface_values = aesara.shared((np.arange(2, 4, dtype=float).reshape(2, -1)),
                                            "Value of the surface to compute")
        self.n_surface_op_float = self.surface_values

        # FAULTS
        # ------
        # Init fault relation matrix
        self.fault_relation = aesara.shared(np.array([[0, 1, 0, 1],
                                                      [0, 0, 1, 1],
                                                      [0, 0, 0, 1],
                                                      [0, 0, 0, 0]]), 'fault relation matrix')

        self.compute_weights = T.vector('Vector controlling if weights must be recomputed', dtype=bool)
        self.compute_scalar = T.vector('Vector controlling if scalar matrix must be recomputed', dtype=bool)
        self.compute_block = T.vector('Vector controlling if block matrix must be recomputed', dtype=bool)

        self.inf_factor = aesara.shared(np.ones(200, dtype='int32') * 10, 'Arbitrary scalar to make df infinite')

        # STRUCTURE
        # ---------
        # This parameters give me the shape of the different groups of data. I pass all data together and I threshold it
        # using these values to the different potential fields and surfaces
        #self.is_fault = is_fault
        #self.is_lith = i
        self.n_faults = aesara.shared(0, 'Number of faults')
        self.n_surfaces_per_series = aesara.shared(np.arange(2, dtype='int32'), 'List with the number of surfaces')
        self.n_universal_eq_T = aesara.shared(np.zeros(5, dtype='int32'), "Grade of the universal drift")



        # This is not accumulative
        self.number_of_points_per_surface_T = aesara.shared(np.zeros(3, dtype='int32'))  # TODO is DEP?
        self.number_of_points_per_surface_T_op = self.number_of_points_per_surface_T
        # This is accumulative
        self.npf = aesara.shared(np.zeros(3, dtype='int32'), 'Number of points per surface accumulative')
        self.npf_op = self.npf[[0, -2]]
        self.len_series_i = aesara.shared(np.arange(2, dtype='int32'), 'Length of surface_points in every series')
        self.len_series_f = aesara.shared(np.arange(2, dtype='int32'), 'Length of foliations in every series')

        self.weights_matrix = aesara.shared(np.zeros((3, 10000)), 'Weights matrix')
        self.scalar_fields_matrix = aesara.shared(np.zeros((3, 10000)), 'Scalar matrix')
        self.block_matrix = aesara.shared(np.zeros((3,10000)), "block matrix")#T.zeros((self.n_surface.shape[0], self.grid_val_T.shape[0] + 2 * self.len_points))
        #T.zeros((self.n_surface.shape[0], self.grid_val_T.shape[0] + 2 * self.len_points))
        if 'initial matrices' in self.verbose:
            self.block_matrix = aesara.printing.Print("block_matrix")(self.block_matrix)
            self.scalar_fields_matrix = aesara.printing.Print("scalar_fields")(self.scalar_fields_matrix)

    def compute_all_series(self):
        # # Change the flag to extend the graph in the compute fault and compute series function
        # lith_matrix = T.zeros((0, 0, self.grid_val_T.shape[0] + 2 * self.len_points))
        #
        # # Init to matrix which contains the block and scalar field of every fault
        # self.fault_matrix = T.zeros((self.n_faults * 2, self.grid_val_T.shape[0] + 2 * self.len_points))
        # self.fault_matrix_f = T.zeros((self.n_faults * 2, self.grid_val_T.shape[0] + 2 * self.len_points))
        #
        # self.final_scalar_field_at_surfaces_op = self.final_scalar_field_at_surfaces
        # self.final_potential_field_at_faults_op = self.final_scalar_field_at_faults
        #
        # # Init df block. Here we store the block and potential field results of one iteration
        # self.fault_block_init = T.zeros((2, self.grid_val_T.shape[0] + 2 * self.len_points))
        # self.fault_block_init.name = 'final block of df init'
        # self.yet_simulated = T.nonzero(T.eq(self.fault_block_init[0, :], 0))[0]



        # Looping
        series, updates1 = aesara.scan(
            fn=self.compute_a_series,
            outputs_info=[
                dict(initial=self.weights_matrix, taps=[0]),
                dict(initial=self.scalar_fields_matrix, taps=[0]),
                dict(initial=self.block_matrix, taps=[0])
                ],  # This line may be used for the df network
            sequences=[dict(input=self.len_series_i, taps=[0, 1]),
                       dict(input=self.len_series_f, taps=[0, 1]),
                       dict(input=self.n_surfaces_per_series, taps=[0, 1]),
                       dict(input=self.n_universal_eq_T, taps=[0]),
                       dict(input=self.compute_weights, taps=[0]),
                       dict(input=self.compute_scalar, taps=[0]),
                       dict(input=self.compute_block, taps=[0])],
           # non_sequences=self.fault_block_init,
            name='Looping',
            return_list=True,
            profile=False
        )

        return series

    def compute_a_series(self,
                         len_i_0, len_i_1,
                         len_f_0, len_f_1,
                         n_form_per_serie_0, n_form_per_serie_1,
                         u_grade_iter,
                         weights_matrix, scalar_fields, block_matrix,
                         compute_weight, compute_scalar, compute_block
                         ):
        """
        Function that loops each fault, generating a potential field for each on them with the respective block model

        Args:
            len_i_0: Lenght of rest of previous series
            len_i_1: Lenght of rest for the computed series
            len_f_0: Lenght of dips of previous series
            len_f_1: Length of dips of the computed series
            n_form_per_serie_0: Number of surfaces of previous series
            n_form_per_serie_1: Number of surfaces of the computed series

        Returns:
            aesara.tensor.matrix: block model derived from the df that afterwards is used as a drift for the "real"
            data
        """

        # THIS IS THE FAULTS BLOCK.
        # ==================
        # Preparing the data
        # ==================

        # compute the youngest fault and consecutively the others

        # aesara shared
        self.number_of_points_per_surface_T_op = self.number_of_points_per_surface_T[n_form_per_serie_0: n_form_per_serie_1]
        self.n_surface_op = self.n_surface[n_form_per_serie_0: n_form_per_serie_1]
        self.n_surface_op_float = self.surface_values[:, n_form_per_serie_0: n_form_per_serie_1 + 1]
        self.npf_op = self.npf[n_form_per_serie_0: n_form_per_serie_1]
        if 'n_surface' in self.verbose:
            self.n_surface_op = aesara.printing.Print('n_surface_fault')(self.n_surface_op)

        self.n_universal_eq_T_op = u_grade_iter

        self.dips_position = self.dips_position_all[len_f_0: len_f_1, :]
        self.dips_position_tiled = T.tile(self.dips_position, (self.n_dimensions, 1))

        # # aesara Var
        # self.dip_angles = self.dip_angles_all[len_f_0: len_f_1]
        # self.azimuth = self.azimuth_all[len_f_0: len_f_1]
        # self.polarity = self.polarity_all[len_f_0: len_f_1]

        self.surface_points_op = self.surface_points[len_i_0: len_i_1, :]

        self.weights_op = self.weights[len_i_0: len_i_1]
        # self.ref_layer_points = self.ref_layer_points_all[len_i_0: len_i_1, :]
        # self.rest_layer_points = self.rest_layer_points_all[len_i_0: len_i_1, :]

        # Updating the interface points involved in the iteration. This is important for the fault drift
        self.len_i_0 = len_i_0
        self.len_i_1 = len_i_1

        if 'lengths' in self.verbose:
            self.len_i_0 = aesara.printing.Print('len_i_0')(self.len_i_0)
            self.len_i_1 = aesara.printing.Print('len_i_1')(self.len_i_1)
            self.len_points = aesara.printing.Print('len_points')(self.len_points)

        # Extracting a the subset of the fault matrix to the scalar field of the current iterations
        faults_relation_op = self.fault_relation[:, T.cast(self.n_surface_op-1, 'int8')]
        faults_relation_rep = T.repeat(faults_relation_op, 1)

        if 'faults_relation' in self.verbose:
            faults_relation_rep = aesara.printing.Print('SELECT')(faults_relation_rep)
        # if len(self.gradients) is not 0:
        #     fault_matrix = block_matrix[::5][T.nonzero(T.cast(faults_relation_rep, "int8"))[0], :]
        # else:

        fault_matrix = block_matrix[T.nonzero(T.cast(faults_relation_rep, "int8"))[0], :]

        if 'fault_matrix_loop' in self.verbose:
            fault_matrix = aesara.printing.Print('self fault matrix')(self.fault_matrix)

        # ================================
        # Computing the fault scalar field
        # ================================
        Z_x = self.Z_x_op(self.dips_position_tiled, self.surface_points,
                          self.fault_matrix[2 * self.len_points:], self.fault_matrix[:2 * self.len_points],
                          self.weights_op, self.grid_val_T)

        series_block = aesara.ifelse.ifelse(self.is_fault[T.cast(self.n_surface_op-1, 'int8')][0],
                                            self.export_fault_block(),
                                            self.export_formation_block(),
                                            name="Is faults condition")

        aux_ind = T.max(self.n_surface_op, 0)

        block_matrix = T.set_subtensor(block_matrix[(aux_ind - 1), :], series_block[0])
        scalar_fields = T.set_subtensor(scalar_fields[(aux_ind-1), :], Z_x)

        return block_matrix, scalar_fields

       #
       #  potential_field_values, faults_matrix = self.block_fault(slope=1000)
       #
       #  # Update the block matrix
       #  final_block = T.set_subtensor(
       #              final_block[0, :],
       #              faults_matrix[0])#T.cast(T.cast(faults_matrix, 'bool'), 'int8'))
       #
       #  # Update the potential field matrix
       # # potential_field_values = self.scalar_field_at_all()
       #
       #  final_block = T.set_subtensor(
       #              final_block[1, :],
       #              potential_field_values)
       #
       #  # Store the potential field at the surface_points
       #  self.final_potential_field_at_faults_op = T.set_subtensor(self.final_potential_field_at_faults_op[self.n_surface_op-1],
       #                                                            self.scalar_field_at_surface_points_values)
       #
       #
       #
       #  if len(self.gradients) is not 0:
       #      weights = self.extend_dual_kriging()
       #      gradients = self.gradient_field_at_all(weights, self.gradients)
       #      final_block = T.set_subtensor(
       #          final_block[2:, :],
       #          gradients)
       #      # Setting the values of the fault matrix computed in the current iteration
       #      fault_matrix = T.set_subtensor(fault_matrix[(aux_ind - 1) * 5:aux_ind * 5, :], final_block)
       #
       #  else:
       #      # Setting the values of the fault matrix computed in the current iteration
       #      fault_matrix = T.set_subtensor(fault_matrix[(aux_ind-1)*2:aux_ind*2, :], final_block)
       #
       #  return fault_matrix, self.final_potential_field_at_faults_op,







#
# class aesaraMask(object):
#     def __init__(self):
#
#         self.Z_x = T.vector('Value of the potential field at every x_to_intep point')
#         self.is_onlap = aesara.shared(np.array([False, False], dtype=bool))
#
#         self.npf = aesara.shared(np.zeros(3, dtype='int32'), 'Number of points per surface accumulative')
#         self.len_points = T.vector('Length of rest or ref surface points arrays')
#
#     def get_scalar_field_at_surface_points(self):
#         self.scalar_field_at_surface_points_values = self.Z_x[-2*self.len_points: -self.len_points][self.npf]
#
#     def set_mask(self):
#
#
#
#         self.mask = T.nonzero(T.le(self.Z_x, self.scalar_field_at_surface_points_values]))[
#             0]  # This -1 comes to get the last scalar field value (the bottom) of the previous series
#         self.mask.name = 'Indices where the scalar field overprints'


class aesaraExportGradient(aesaraExport):

    def __init__(self):

        super().__init__()

    def contribution_interface_gradient(self, direction='x', grid_val=None, weights=None):
        """
        Computation of the contribution of the foliations at every point to interpolate

        Returns:
            aesara.tensor.vector: Contribution of all foliations (input) at every point to interpolate
        """

        if direction == 'x':
            dir_val = 0
        elif direction == 'y':
            dir_val = 1
        elif direction == 'z':
            dir_val = 2
        else:
            raise AttributeError('Directions muxt be x, y or z')

        if weights is None:
            weights = self.extend_dual_kriging()
        if grid_val is None:
            grid_val = self.x_to_interpolate()

        length_of_CG, length_of_CGI = self.matrices_shapes()[:2]

        # Cartesian distances between the point to simulate and the dips
        hu_rest = (- self.rest_layer_points[:, dir_val] + grid_val[:, dir_val].reshape(
            (grid_val[:, dir_val].shape[0], 1)))
        hu_ref = (- self.ref_layer_points[:, dir_val] + grid_val[:, dir_val].reshape(
            (grid_val[:, dir_val].shape[0], 1)))

        # Euclidian distances

        sed_grid_rest = self.squared_euclidean_distances(grid_val, self.rest_layer_points)
        sed_grid_ref = self.squared_euclidean_distances(grid_val, self.ref_layer_points)

        # Gradient contribution
        self.gi_reescale = 2

        sigma_0_grad = T.sum(
            (weights[length_of_CG:length_of_CG + length_of_CGI] *
             self.gi_reescale * (
                     (hu_rest *
                      (sed_grid_rest < self.a_T) *  # first derivative
                      (- self.c_o_T * ((-14 / self.a_T ** 2) + 105 / 4 * sed_grid_rest / self.a_T ** 3 -
                                       35 / 2 * sed_grid_rest ** 3 / self.a_T ** 5 +
                                       21 / 4 * sed_grid_rest ** 5 / self.a_T ** 7))) -
                     (hu_ref *
                      (sed_grid_ref < self.a_T) *  # first derivative
                      (- self.c_o_T * ((-14 / self.a_T ** 2) + 105 / 4 * sed_grid_ref / self.a_T ** 3 -
                                       35 / 2 * sed_grid_ref ** 3 / self.a_T ** 5 +
                                       21 / 4 * sed_grid_ref ** 5 / self.a_T ** 7)))).T),
            axis=0)

        return sigma_0_grad

    def contribution_gradient(self, direction='x', grid_val=None, weights=None):

        if direction == 'x':
            direction_val = 0
        if direction == 'y':
            direction_val = 1
        if direction == 'z':
            direction_val = 2
        self.gi_reescale = aesara.shared(1)

        if weights is None:
            weights = self.extend_dual_kriging()
        if grid_val is None:
            grid_val = self.x_to_interpolate()

        length_of_CG = self.matrices_shapes()[0]

        # Cartesian distances between the point to simulate and the dips
        # TODO optimize to compute this only once?
        # Euclidean distances
        sed_dips_SimPoint = self.squared_euclidean_distances(grid_val, self.dips_position_tiled).T

        if 'sed_dips_SimPoint' in self.verbose:
            sed_dips_SimPoint = aesara.printing.Print('sed_dips_SimPoint')(sed_dips_SimPoint)

        # Cartesian distances between dips positions
        h_u = T.tile(self.dips_position[:, direction_val] - grid_val[:, direction_val].reshape(
            (grid_val[:, direction_val].shape[0], 1)), 3)
        h_v = T.horizontal_stack(
            T.tile(self.dips_position[:, 0] - grid_val[:, 0].reshape((grid_val[:, 0].shape[0], 1)),
                   1),
            T.tile(self.dips_position[:, 1] - grid_val[:, 1].reshape((grid_val[:, 1].shape[0], 1)),
                   1),
            T.tile(self.dips_position[:, 2] - grid_val[:, 2].reshape((grid_val[:, 2].shape[0], 1)),
                   1))

        perpendicularity_vector = T.zeros(T.stack([length_of_CG], axis=0))
        perpendicularity_vector = T.set_subtensor(
            perpendicularity_vector[
            self.dips_position.shape[0] * direction_val:self.dips_position.shape[0] * (direction_val + 1)], 1)

        sigma_0_grad = T.sum(
            (weights[:length_of_CG] * (
                    ((-h_u * h_v).T / sed_dips_SimPoint ** 2) *
                    ((
                             (sed_dips_SimPoint < self.a_T) *  # first derivative
                             (-self.c_o_T * ((-14 / self.a_T ** 2) + 105 / 4 * sed_dips_SimPoint / self.a_T ** 3 -
                                             35 / 2 * sed_dips_SimPoint ** 3 / self.a_T ** 5 +
                                             21 / 4 * sed_dips_SimPoint ** 5 / self.a_T ** 7))) +
                     (sed_dips_SimPoint < self.a_T) *  # Second derivative
                     self.c_o_T * 7 * (9 * sed_dips_SimPoint ** 5 - 20 * self.a_T ** 2 * sed_dips_SimPoint ** 3 +
                                       15 * self.a_T ** 4 * sed_dips_SimPoint - 4 * self.a_T ** 5) / (
                             2 * self.a_T ** 7)) -
                    (perpendicularity_vector.reshape((-1, 1)) *
                     ((sed_dips_SimPoint < self.a_T) *  # first derivative
                      self.c_o_T * ((-14 / self.a_T ** 2) + 105 / 4 * sed_dips_SimPoint / self.a_T ** 3 -
                                    35 / 2 * sed_dips_SimPoint ** 3 / self.a_T ** 5 +
                                    21 / 4 * sed_dips_SimPoint ** 5 / self.a_T ** 7)))

            ))
            , axis=0)

        return sigma_0_grad

    def contribution_universal_drift_d(self, direction='x', grid_val=None, weights=None, a=0, b=100000000):
        if weights is None:
            weights = self.extend_dual_kriging()
        if grid_val is None:
            grid_val = self.x_to_interpolate()

        self.gi_reescale = aesara.shared(2)

        length_of_CG, length_of_CGI, length_of_U_I, length_of_faults, length_of_C = self.matrices_shapes()

        # These are the magic terms to get the same as geomodeller
        i_rescale_aux = T.repeat(self.gi_reescale, 9)
        i_rescale_aux = T.set_subtensor(i_rescale_aux[:3], 1)
        _aux_magic_term = T.tile(i_rescale_aux[:self.n_universal_eq_T_op], (grid_val.shape[0], 1))

        n = self.dips_position.shape[0]
        n = grid_val.shape[0]
        U_G = T.zeros((n, 9))

        if direction == 'x':
            # x
            U_G = T.set_subtensor(U_G[:, 0], 1)
            # x**2
            U_G = T.set_subtensor(U_G[:, 3], 2 * self.gi_reescale * grid_val[:, 0])

            # xy
            U_G = T.set_subtensor(U_G[:, 6], self.gi_reescale * grid_val[:, 1])  # This is y
            # xz
            U_G = T.set_subtensor(U_G[:, 7], self.gi_reescale * grid_val[:, 2])  # This is z

        if direction == 'y':
            # y
            U_G = T.set_subtensor(U_G[:, 1], 1)
            # y**2
            U_G = T.set_subtensor(U_G[:, 4], 2 * self.gi_reescale * grid_val[:, 1])
            # xy
            U_G = T.set_subtensor(U_G[:, 6], self.gi_reescale * grid_val[:, 0])  # This is x
            # yz
            U_G = T.set_subtensor(U_G[:, 8], self.gi_reescale * grid_val[:, 2])  # This is z

        if direction == 'z':
            # z
            U_G = T.set_subtensor(U_G[:, 2], 1)

            # z**2
            U_G = T.set_subtensor(U_G[:, 5], 2 * self.gi_reescale * grid_val[:, 2])

            # xz
            U_G = T.set_subtensor(U_G[:, 7], self.gi_reescale * grid_val[:, 0])  # This is x

            # yz
            U_G = T.set_subtensor(U_G[:, 8], self.gi_reescale * grid_val[:, 1])  # This is y

        # Drif contribution
        f_0 = (T.sum(
            weights[
            length_of_CG + length_of_CGI:length_of_CG + length_of_CGI + length_of_U_I] *
            U_G.T[:self.n_universal_eq_T_op]
            , axis=0))

        return f_0

    def gradient_field_loop_x(self, a, b, Z_x, grid_val, weights, val):
        direction = 'x'
        sigma_0_grad = self.contribution_gradient(direction, grid_val[a:b], weights[:, a:b])
        sigma_0_interf_gradient = self.contribution_interface_gradient(direction, grid_val[a:b], weights[:, a:b])
        f_0 = self.contribution_universal_drift_d(direction, grid_val[a:b], weights[:, a:b], a, b)
        #f_1 = self.faults_contribution(weights[:, a:b], a, b)

        # Add an arbitrary number at the potential field to get unique values for each of them
        partial_Z_x = (sigma_0_grad + sigma_0_interf_gradient + f_0)

        Z_x = T.set_subtensor(Z_x[a:b], partial_Z_x)

        return Z_x

    def gradient_field_loop_y(self, a, b, Z_x, grid_val, weights, val):
        direction = 'y'
        sigma_0_grad = self.contribution_gradient(direction, grid_val[a:b], weights[:, a:b])
        sigma_0_interf_gradient = self.contribution_interface_gradient(direction, grid_val[a:b], weights[:, a:b])
        f_0 = self.contribution_universal_drift_d(direction, grid_val[a:b], weights[:, a:b], a, b)
        #f_1 = self.faults_contribution(weights[:, a:b], a, b)

        # Add an arbitrary number at the potential field to get unique values for each of them
        partial_Z_x = (sigma_0_grad + sigma_0_interf_gradient + f_0)

        Z_x = T.set_subtensor(Z_x[a:b], partial_Z_x)

        return Z_x

    def gradient_field_loop_z(self, a, b, Z_x, grid_val, weights, val):
        direction = 'z'
        sigma_0_grad = self.contribution_gradient(direction, grid_val[a:b], weights[:, a:b])
        sigma_0_interf_gradient = self.contribution_interface_gradient(direction, grid_val[a:b], weights[:, a:b])
        f_0 = self.contribution_universal_drift_d(direction, grid_val[a:b], weights[:, a:b], a, b)
        #f_1 = self.faults_contribution(weights[:, a:b], a, b)

        # Add an arbitrary number at the potential field to get unique values for each of them
        partial_Z_x = (sigma_0_grad + sigma_0_interf_gradient + f_0)

        Z_x = T.set_subtensor(Z_x[a:b], partial_Z_x)

        return Z_x

    def gradient_field_at_all(self, weights=None, gradients=[]):

        grid_val = self.x_to_interpolate()
        if weights is None:
            weights = self.extend_dual_kriging()

        grid_shape = T.stack([grid_val.shape[0]], axis=0)

        Z_x_init = T.zeros(grid_shape, dtype='float32')
        if 'grid_shape' in self.verbose:
            grid_shape = aesara.printing.Print('grid_shape')(grid_shape)

        steps = 1e13 / self.matrices_shapes()[-1] / grid_shape
        slices = T.concatenate((T.arange(0, grid_shape[0], steps[0], dtype='int64'), grid_shape))

        if 'slices' in self.verbose:
            slices = aesara.printing.Print('slices')(slices)

        G_field = T.zeros((3, self.grid_val_T.shape[0]))

        if 'Gx' in gradients:
            Gx_loop, updates5 = aesara.scan(
                fn=self.gradient_field_loop_x,
                outputs_info=[Z_x_init],
                sequences=[dict(input=slices, taps=[0, 1])],
                non_sequences=[grid_val, weights, self.n_surface_op],
                profile=False,
                name='Looping grid x',
                return_list=True)

            Gx = Gx_loop[-1][-1]
            Gx.name = 'Value of the gradient field X at every point'
            G_field = T.set_subtensor(G_field[0, :], Gx)

        if 'Gy' in gradients:
            Gy_loop, updates6 = aesara.scan(
                fn=self.gradient_field_loop_y,
                outputs_info=[Z_x_init],
                sequences=[dict(input=slices, taps=[0, 1])],
                non_sequences=[grid_val, weights, self.n_surface_op],
                profile=False,
                name='Looping grid y',
                return_list=True)

            Gy = Gy_loop[-1][-1]
            Gy.name = 'Value of the gradient field X at every point'
            G_field = T.set_subtensor(G_field[1, :], Gy)

        if 'Gz' in gradients:
            Gz_loop, updates7 = aesara.scan(
                fn=self.gradient_field_loop_z,
                outputs_info=[Z_x_init],
                sequences=[dict(input=slices, taps=[0, 1])],
                non_sequences=[grid_val, weights, self.n_surface_op],
                profile=False,
                name='Looping grid z',
                return_list=True)

            Gz = Gz_loop[-1][-1]
            Gz.name = 'Value of the gradient field X at every point'
            G_field = T.set_subtensor(G_field[2, :], Gz)

        if str(sys._getframe().f_code.co_name) in self.verbose:
            Z_x = aesara.printing.Print('Potential field at all points')(Z_x)

        return G_field


"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.

Tested on Ubuntu 16

Created on 23/06/2018

@author: Miguel de la Varga, Alexander Schaaf
"""
import numpy as np
import pandas as pn


def find_surface_points_from_block(block, value):
    """
    TODO: Is this dep by find_interfaces_from_block_bottoms?
    Find the voxel at an interface. We shift left since gempy is based on bottoms

    Args:
        block (ndarray):
        value:

    Returns:

    """
    A = block > value
    # Matrix shifting along axis
    B = A  #
    x_shift = B[:-1, :, :] ^ B[1:, :, :]

    # Matrix shifting along axis
    y_shift = B[:, :-1, :] ^ B[:, 1:, :]

    # Matrix shifting along axis
    z_shift = B[:, :, :-1] ^ B[:, :, 1:]

    final_bool = np.zeros_like(block, dtype=bool)
    final_bool[:-1, :-1, :-1] = x_shift[:, :-1, :-1] + y_shift[:-1, :, :-1] + z_shift[-1:, -1:, :]

    return final_bool


def find_interfaces_from_block_bottoms(block, value, shift=2):
    """
    Find the voxel at an interface. We shift left since gempy is based on bottoms

    Args:
        block (ndarray): matrix with the scalar values
        value: value of which you are looking the interfaces
        shift (int): Number of elements shifted

    Returns:

    """
    A = block == value
    final_bool = np.zeros_like(block, dtype=bool)

    # Matrix shifting along axis 0
    x_shift = A[:-shift, :, :] ^ A[shift:, :, :]

    # Matrix shifting along axis 1
    y_shift = A[:, :-shift, :] ^ A[:, shift:, :]

    # Matrix shifting along axis 2
    z_shift = A[:, :, :-shift] ^ A[:, :, shift:]
    final_bool[shift:, shift:, shift:] = (x_shift[:, shift:, shift:] +
                                          y_shift[shift:, :, shift:] +
                                          z_shift[shift:, shift:, :])
    return final_bool


def surface_points_from_surface_points_block(block_bool, block_grid, formation='default_formation', series='Default_series',
                                     formation_number=1, order_series=1, n_points=20):

    assert np.ravel(block_bool).shape[0] == block_grid.shape[0], 'Grid and block block must have the same size. If you' \
                                                           'are importing a model from noddy make sure that the' \
                                                           'resolution is the same'
    coord_select = block_grid[np.ravel(block_bool)]

    loc_points = np.linspace(0, coord_select.shape[0]-1, n_points, dtype=int)

    # Init dataframe
    p = pn.DataFrame(columns=['X', 'Y', 'Z', 'formation', 'series', 'formation_number',
                              'order_series', 'isFault'])

    p[['X', 'Y', 'Z']] = pn.DataFrame(coord_select[loc_points])
    p['formation'] = formation
    p['series'] = series
    p['formation_number'] = formation_number
    p['order_series'] = order_series

    return p


def set_surface_points_from_block(geo_data, block, block_grid=None, n_points=20, reset_index=False):
    values = np.unique(np.round(block))
    values.sort()
    values = values[:-1]

    if block_grid is None:
        block_grid = geo_data._grid.values

    for e, value in enumerate(values):
        block_bool = find_surface_points_from_block(block, value)

        geo_data.set_interface_object(surface_points_from_surface_points_block(block_bool, block_grid,
                                                                       formation='formation_'+str(e), series='Default_series',
                                                                       formation_number=e, order_series=1,
                                                                       n_points=n_points), append=True)
        if reset_index:
            geo_data._surface_points.reset_index(drop=True, inplace=True)

    return geo_data


class VanMisesFisher:
    def __init__(self, mu, kappa, dim=3):
        """van Mises-Fisher distribution for sampling vector components from n-dimensional spheres.

        Adapted from source: https://github.com/pymc-devs/pymc3/issues/2458
        
        a von Mises-Fisher distribution can be fit using scipy.stats.vonmises

        Args:
            mu (np.ndarray): Mean direction of vector [Gx, Gy, Gz]
            kappa (float): Concentration parameter (the lower the higher the spread on the sphere)
            dim (int, optional): Dimensionality of the Sphere
        """
        self.mu = mu
        self.kappa = kappa
        self.dim = dim

    def rvs(self, n=1):
        """Obtain n samples from van Mises-Fisher distribution.

        Args:
            n (int): Number of samples to draw

        Returns:
            np.ndarray with shape (n, 3) containing samples.

        """
        result = np.zeros((n, self.dim))
        for nn in range(n):
            # sample offset from center (on sphere) with spread kappa
            w = self._sample_weight()
            # sample a point v on the unit sphere that's orthogonal to mu
            v = self._sample_orthonormal_to()
            # compute new point
            result[nn, :] = v * np.sqrt(1. - w** 2) + w * self.mu
        return result

    def _sample_weight(self):
        """Who likes documentation anyways. This is totally intuitive and trivial."""
        dim = self.dim - 1  # since S^{n-1}
        b = dim / (np.sqrt(4. * self.kappa ** 2 + dim ** 2) + 2 * self.kappa)
        x = (1. - b) / (1. + b)
        c = self.kappa * x + dim * np.log(1 - x ** 2)

        while True:
            z = np.random.beta(dim / 2., dim / 2.)
            w = (1. - (1. + b) * z) / (1. - (1. - b) * z)
            u = np.random.uniform(low=0, high=1)
            if self.kappa * w + dim * np.log(1. - x * w) - c >= np.log(u):
                return w

    def _sample_orthonormal_to(self):
        """Who likes documentation anyways. This is totally intuitive and trivial."""
        v = np.random.randn(self.mu.shape[0])
        proj_mu_v = self.mu * np.dot(self.mu, v) / np.linalg.norm(self.mu)
        orthto = v - proj_mu_v
        return orthto / np.linalg.norm(orthto)

    def stats(self):
        return self.mu, self.kappa


def change_data(interp_data, geo_data_stoch, priors):
    """Changes input data with prior distributions (scipy.stats distributions) given in list.
    Prior distribution objects must contain .rvs() method for drawing samples.

    Args:
        interp_data:
        geo_data_stoch:
        priors:
        verbose:

    Returns:

    """
    prior_draws = []
    for prior in priors:
        if hasattr(prior, "gradient"):
            value = prior.rvs()
        else:
            value = prior.rvs() / interp_data.rescaling_factor
        prior_draws.append(value)

        if prior.index_interf is not None:
            if prior.replace:  # replace the value
                # geo_data.interfaces.set_value(prior.index_interf, prior.column, prior.rvs() / rf)
                interp_data.geo_data_res.interfaces.loc[prior.index_interf, prior.column] = value
            else:  # add value
                interp_data.geo_data_res.interfaces.loc[prior.index_interf, prior.column] = geo_data_stoch.interfaces.loc[
                                                                                prior.index_interf, prior.column] + value
        if prior.index_orient is not None:
            if prior.replace:  # replace the value
                interp_data.geo_data_res._orientations.loc[prior.index_orient, prior.column] = value
            else:  # add value
                interp_data.geo_data_res._orientations.loc[prior.index_orient, prior.column] = geo_data_stoch._orientations.loc[
                                                                                  prior.index_orient, prior.column] + value
    return prior_draws



"""
Claudius
~~~~~~~~

"""

# %%
import sys, os
os.environ["aesara_FLAGS"] = "mode=FAST_RUN,device=cpu"

# Importing gempy
import gempy_legacy as gp

# Aux imports
import numpy as np
import pandas as pn

# %%
# Loading data from repository:
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# 
# With pandas we can do it directly from the web and with the right args
# we can directly tidy the data in gempy style:
# 

# %% 
dfs = []
for letter in 'ABCD':
    dfs.append(pn.read_csv('https://raw.githubusercontent.com/Loop3D/ImplicitBenchmark/master/Claudius/' +
                           letter + 'Points.csv', sep=';',
                           names=['X', 'Y', 'Z', 'surface', 'cutoff'], header=0)[::5])
# Add fault:
dfs.append(pn.read_csv('https://raw.githubusercontent.com/Loop3D/ImplicitBenchmark/master/Claudius/Fault.csv',
                       names=['X', 'Y', 'Z', 'surface'], header=0, sep=';'))

surface_points = pn.concat(dfs, sort=True)
surface_points['surface'] =surface_points['surface'].astype('str')
# surface_points['surface'] = surface_points['surface'].astype('str')
surface_points.reset_index(inplace=True, drop=False)
surface_points.tail()

# %%
surface_points.dtypes

# %%
# How many points are per surface
# 

# %% 
surface_points.groupby('surface').count()

# %%
# Now we do the same with the orientations:
# 

# %% 
dfs = []

for surf in ['0', '330']:
    o = pn.read_csv('https://raw.githubusercontent.com/Loop3D/ImplicitBenchmark/master/Claudius/Dips.csv', sep=';',
                    names=['X', 'Y', 'Z', 'G_x', 'G_y', 'G_z', '-'], header=1)

    # Orientation needs to belong to a surface. This is mainly to categorize to which series belong and to
    # use the same color
    o['surface'] = surf
    dfs.append(o)
orientations = pn.concat(dfs, sort=True)
orientations.reset_index(inplace=True, drop=False)

orientations.tail()

# %%
orientations.dtypes

# %%
# Data initialization:
# ~~~~~~~~~~~~~~~~~~~~
# 
# Suggested size of the axis-aligned modeling box: Origin: 548800 7816600
# -8400 Maximum: 552500 7822000 -11010
# 
# Suggested resolution: 100m x 100m x -90m (grid size 38 x 55 x 30)
# 

# %% 
# Number of voxels:
np.array([38, 55, 30]).prod()

# %% 
geo_model = gp.create_model('Claudius')
# Importing the data from csv files and settign extent and resolution
geo_model = gp.init_data(geo_model,
                         extent=[548800, 552500, 7816600, 7822000, -11010, -8400], resolution=[38, 55, 30],
                         surface_points_df=surface_points[::5], orientations_df=orientations, surface_name='surface',
                         add_basement=True)

# %%
# We are going to increase the smoothness (nugget) of the data to increase
# the conditional number of the matrix:
# 

# %% 
geo_model.modify_surface_points(geo_model.surface_points.df.index, smooth=0.1).df.tail()

# %%
# Also the original poles are pointing downwards. We can change the
# direction by calling the following:
# 

# %% 
geo_model.modify_orientations(geo_model.orientations.df.index, polarity=-1).df.tail()

# %%
# We need an orientation per series/fault. The faults does not have
# orientation so the easiest is to create an orientation from the surface
# points availablle:
# 

# %% 
fault_idx = geo_model.surface_points.df.index[geo_model.surface_points.df['surface'] == 'Claudius_fault']
gp.set_orientation_from_surface_points(geo_model, fault_idx).df.tail()

# %%
# Now we can see how the data looks so far:
# 

# %%
geo_model.surfaces

# %% 
gp.plot_2d(geo_model, direction='y')

# %%
# By default all surfaces belong to one unique series.
# 

# %% 
geo_model.surfaces

# %%
# We will need to separate with surface belong to each series:
# 

# %% 
stratigraphy = 'fixed'

# %% 
if stratigraphy == 'original':
    gp.map_stack_to_surfaces(geo_model, {'Fault': 'Claudius_fault',
                                         'Default series': ('0', '60', '250', '330'),
                                         })
    # Ordering the events from younger to older:
    geo_model.reorder_series(['Fault', 'Default series', 'Basement'])


elif stratigraphy == 'fixed':
    gp.map_stack_to_surfaces(geo_model, {'Default series': ('0', '60', '250'),
                                         'Fault': 'Claudius_fault',
                                         'Uncomformity': '330',
                                         })
    # Ordering the events from younger to older:
    geo_model.reorder_series(['Default series', 'Fault', 'Uncomformity', 'Basement'])

# %%
# So far we did not specify which series/faults are actula faults:
# 

# %% 
geo_model.set_is_fault('Fault')

# %%
# Ordering the events from younger to older:
# 

# %% 
# geo_model.reorder_series(['Default series', 'Fault', 'Uncomformity', 'Basement'])


# %%
# Check which series/faults are affected by other faults (rows offset
# columns):
# 

# %% 
geo_model.faults.faults_relations_df

# %%
# Now we are good to go:
# 

# %% 
gp.set_interpolator(geo_model, aesara_optimizer='fast_run',
                    compile_aesara=True)

# %% 
gp.compute_model(geo_model)

# %% 
sect = [35]

gp.plot_2d(geo_model, cell_number=sect, series_n=1, show_scalar=True, direction='x')


# %% 
gp.plot_2d(geo_model, cell_number=sect, show_data=True, direction='x')

# %% 
gp.plot_2d(geo_model, cell_number=[28], series_n=0, direction='y', show_scalar=True)
gp.plot_2d(geo_model, cell_number=[28], series_n=1, direction='y', show_scalar=True)
gp.plot_2d(geo_model, cell_number=[28], series_n=2, direction='y', show_scalar=True)

# %% 
gp.plot_2d(geo_model, cell_number=[28], show_data=True, direction='y')

# %%

# sphinx_gallery_thumbnail_number = 8
gp.plot_3d(geo_model)



# %%
# Export data:
# ~~~~~~~~~~~~
# 
# The solution is stored in a numpy array of the following shape. Axis 0
# are the scalar fields of each correspondent series/faults in the
# following order (except basement):
# 

# %% 
geo_model.series

# %%
# For the surfaces, there are two numpy arrays, one with vertices and the
# other with triangles. Axis 0 is each surface in the order:
# 

# %% 
geo_model.surfaces


# %%
# np.save('Claudius_scalar', geo_model.solutions.scalar_field_matrix)
# np.save('Claudius_ver', geo_model.solutions.vertices)
# np.save('Claudius_edges', geo_model.solutions.edges)
# gp.plot.export_to_vtk(geo_model, 'Claudius')


# %%
# Timing:
# -------
# 
# Fault
# ~~~~~
# 
# Dense 20k input, 62k voxels
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^
# 
# -  CPU Memory 8 Gb 44.9 s ± 150 ms per loop (mean ± std. dev. of 7 runs,
#    1 loop each)
# -  GPU Memory 6.8 gb:
# 
#    -  2.13 s ± 3.39 ms per loop (mean ± std. dev. of 7 runs, 1 loop
#       each) + steps **str** = [64.56394268] + steps **str** =
#       [9927.69441126] + steps **str** = [196.15202667]
# 
#    -  1.13 s ± 2.08 ms per loop (mean ± std. dev. of 7 runs, 1 loop
#       each)
# 
#       ::
# 
#          + steps __str__ = [645.63943742]
#          + steps __str__ = [99276.94573919]
#          + steps __str__ = [1961.52029888]
# 


# %%
# Export to gocad
# ---------------
# 

# %% 
def write_property_to_gocad_voxet(propertyfilename, propertyvalues):
    """
    This function writes a numpy array into the right format for a gocad
    voxet property file. This assumet there is a property already added to the .vo file,
    and is just updating the file.
    propertyfile - string giving the path to the file to write
    propertyvalues - numpy array nz,ny,nx ordering and in float format
    """
    propertyvalues = propertyvalues.astype('>f4')  # big endian
    #     array = propertyvalues.newbyteorder()
    propertyvalues.tofile(propertyfilename)


# %%
write_property_to_gocad_voxet('claudius_sf_gempy',
                              geo_model.solutions.scalar_field_matrix[1].reshape([38, 55, 30]).ravel('F'))

gp.save_model(geo_model)


import numpy as np
from typing import Union
import warnings
from skimage import measure

from ..utils.input_manipulation import find_interfaces_from_block_bottoms
from .surfaces import Surfaces
from .grid import Grid
from .data_modules.stack import Series, Stack
from ..utils.meta import _setdoc_pro
import gempy_legacy.utils.docstring as ds

try:
    from gempy_legacy.core.xsolution import XSolution

    xsolution_imported = True
    inheritance = XSolution
except ImportError:
    inheritance = object
    xsolution_imported = False


@_setdoc_pro(
    [Grid.__doc__, Surfaces.__doc__, Series.__doc__, ds.weights_vector, ds.sfai,
     ds.bai, ds.mai, ds.vai,
     ds.lith_block, ds.sfm, ds.bm, ds.mm, ds.vm, ds.vertices, ds.edges,
     ds.geological_map])
class Solution(inheritance):
    """This class stores the output of the interpolation and the necessary objects
    to visualize and manipulate this data.

    Depending on the activated grid (see :class:`Grid`) a different number of
     properties are returned returned:

    Args:
        grid (Grid): [s0]
        surfaces (Surfaces): [s1]
        series (Series): [s2]

    Attributes:
        grid (Grid)
        surfaces (Surfaces)
        series (Series)
        weights_vector (numpy.array): [s3]
        scalar_field_at_surface_points (numpy.array): [s4]
        block_at_surface_points (numpy.array): [s5]
        mask_at_surface_points (numpy.array): [s6]
        values_at_surface_points (numpy.array): [s7]
        lith_block (numpy.array): [s8]
        scalar_field_matrix (numpy.array): [s9]
        block_matrix (numpy.array): [s10]
        mask_matrix (numpy.array): [s11]
        mask_matrix_pad (numpy.array): mask matrix padded 2 block in order to guarantee that the layers intersect each
         other after marching cubes
        values_matrix (numpy.array): [s12]
        vertices (list[numpy.array]): [s13]
        edges (list[numpy.array]): [s14]
        geological_map (numpy.array): [s15]

    """

    def __init__(self, grid: Grid = None,
                 surfaces: Surfaces = None,
                 series: Series = None,
                 ):

        if xsolution_imported:
            super().__init__(grid, surfaces, series)

        self.grid = grid
        self.stack = series
        self.surfaces = surfaces

        # Input data results
        self.weights_vector = np.empty(0)
        self.scalar_field_at_surface_points = np.array([])
        self.block_at_surface_points = np.array([])
        self.mask_at_surface_points = np.array([])
        self.values_at_surface_points = np.array([])

        # Regular Grid
        self.lith_block = np.empty(0)
        self.scalar_field_matrix = np.array([])
        self.block_matrix = np.array([])
        self.mask_matrix = np.array([])
        self.mask_matrix_pad = []
        self.values_matrix = np.array([])
        self.gradient = np.empty(0)

        self.vertices = []
        self.edges = []

        self.topology_edges = None
        self.topology_count = None

        self.geological_map = None
        self.sections = None
        self.custom = None

        # Center Grid
        self.fw_gravity = None
        self.fw_magnetics = None

    def __repr__(self):
        return '\nLithology ids \n  %s \n' \
            % (np.array2string(self.lith_block))

    def set_solution_to_regular_grid(self, values: Union[list, np.ndarray],
                                     compute_mesh: bool = True,
                                     compute_mesh_options: dict = None):
        """If regular grid is active set all the solution objects dependent on them and compute mesh.

        Args:
            values (list[np.array]): list with result of the aesara evaluation (values returned by
             :func:`gempy.compute_model` function):

                 - block_matrix
                 - weights_vector
                 - scalar_field_matrix
                 - scalar field at interfaces
                 - mask_matrix

            compute_mesh (bool): if True perform marching cubes algorithm to recover the surface mesh from the
             implicit model.
            compute_mesh_options (dict): options for the marching cube function.
                - rescale: True

        Returns:
            :class:`gempy.core.solutions.Solutions`

        """
        if compute_mesh_options is None:
            compute_mesh_options = {}
        self.set_values_to_regular_grid(values)
        if compute_mesh is True:
            self.compute_all_surfaces(**compute_mesh_options)

        return self

    def set_solutions(self, sol, compute_mesh, sort_surfaces, to_subsurface=False, **kwargs):

        # Set geology:
        self.set_values_to_surface_points(sol)

        if self.grid.active_grids[0] is np.True_:
            self.set_solution_to_regular_grid(sol,
                                              compute_mesh=compute_mesh,
                                              **kwargs)
        if self.grid.active_grids[1] is np.True_:
            self.set_solution_to_custom(sol)
        if self.grid.active_grids[2] is np.True_:
            self.set_solution_to_topography(sol)
        if self.grid.active_grids[3] is np.True_:
            self.set_solution_to_sections(sol)

        self.topology_edges = sol[10]
        self.topology_count = sol[11]

        # Set gravity
        self.fw_gravity = sol[12]

        # TODO: [X] Set magnetcs and [ ] set topology @A.Schaaf probably it should
        #  populate the topology object?
        self.fw_magnetics = sol[13]

        if xsolution_imported and to_subsurface is True:
            self.set_values(sol)
            if compute_mesh is True:
                try:
                    self.set_meshes(self.surfaces)
                except ValueError:
                    warnings.warn('Meshes are empty.')
        return self

    def set_solution_to_custom(self, values: Union[list, np.ndarray]):
        l0, l1 = self.grid.get_grid_args('custom')
        self.custom = np.stack(  # * This has been changed recently. It could give problems how we slice the field
            [values[0][:, l0: l1],
             values[4][:, l0: l1].astype(float)]
        )

    def set_solution_to_topography(self, values: Union[list, np.ndarray]):
        l0, l1 = self.grid.get_grid_args('topography')
        self.geological_map = np.vstack(  # * This has been changed recently. It could give problems how we slice the field
            [values[0][:, l0: l1],
             values[4][:, l0: l1].astype(float)]
        )

    def set_solution_to_sections(self, values: Union[list, np.ndarray]):
        l0, l1 = self.grid.get_grid_args('sections')
        self.sections = np.vstack(  # * This has been changed recently. It could give problems how we slice the field
            [values[0][:, l0: l1],
             values[4][:, l0: l1].astype(float)]
        )

    def set_values_to_regular_grid(self, values: Union[list, np.ndarray]):
        """Set all solution values to the correspondent attribute.

        Args:
            values (np.ndarray): values returned by `function: gempy.compute_model` function
            compute_mesh (bool): if true compute automatically the grid

        Returns:
            :class:`gempy.core.solutions.Solutions`

        """
        regular_grid_length_l0, regular_grid_length_l1 = self.grid.get_grid_args('regular')

        # Lithology final block
        self.lith_block = values[0][0, regular_grid_length_l0: regular_grid_length_l1]

        # Properties
        self.values_matrix = values[0][1:, regular_grid_length_l0: regular_grid_length_l1]

        # Axis 0 is the series. Axis 1 is the value

        self.block_matrix = values[1][:, :, regular_grid_length_l0: regular_grid_length_l1]
        self.fault_block = values[2]
        #self.block_matrix = values[1][:, :,
        #                    regular_grid_length_l0: regular_grid_length_l1]

        #self.fault_block = values[2][0,
        #                  regular_grid_length_l0: regular_grid_length_l1]

        # This here does not make any sense
        self.weights_vector = values[3]
        self.scalar_field_matrix = values[4][:, regular_grid_length_l0: regular_grid_length_l1]
        self.mask_matrix = values[6][:, regular_grid_length_l0: regular_grid_length_l1]
        self.fault_mask = values[7][:, regular_grid_length_l0: regular_grid_length_l1]

        # TODO add topology solutions

        return self

    def set_values_to_surface_points(self, values):
        x_to_intep_length = self.grid.length[-1]
        self.scalar_field_at_surface_points = values[5]
        self.values_at_surface_points = values[0][1:, x_to_intep_length:]
        self.block_at_surface_points = values[1][:, :, x_to_intep_length:]
        # todo disambiguate below from self.scalar_field_at_surface_points
        self._scalar_field_at_surface = values[4][:, x_to_intep_length:]
        self.mask_at_surface_points = values[6][:, x_to_intep_length:]
        return self.scalar_field_at_surface_points

    def compute_marching_cubes_regular_grid(self, level: float, scalar_field,
                                            mask_array=None,
                                            rescale=False, **kwargs):
        """Compute the surface (vertices and edges) of a given surface by computing
         marching cubes (by skimage)

        Args:
            level (float): value of the scalar field at the surface
            scalar_field (np.array): scalar_field vector objects
            mask_array (np.array): mask vector with trues where marching cubes has to be performed
            rescale (bool): if True surfaces will be located between 0 and 1
            **kwargs: skimage.measure.marching_cubes_lewiner args (see below)

        Returns:
            list: vertices, simplices, normals, values

        See Also:

            :func:`skimage.measure.marching_cubes`

        """
        rg = self.grid.regular_grid
        spacing = self.grid.regular_grid.get_dx_dy_dz(rescale=rescale)
        vertices, simplices, normals, values = measure.marching_cubes(
            scalar_field.reshape(rg.resolution),
            level,
            spacing=spacing,
            mask=mask_array, **kwargs)
        idx = [0, 2, 4]
        loc_0 = rg.extent_r[idx] if rescale else rg.extent[idx]
        loc_0 = loc_0 + np.array(spacing) / 2
        vertices += np.array(loc_0).reshape(1, 3)

        return [vertices, simplices, normals, values]

    def padding_mask_matrix(self, mask_topography=True, shift=2):
        """Pad as many elements as in shift to the masking arrays. This is done
         to guarantee intersection of layers if masked marching cubes are done

        Args:
            mask_topography (bool): if True mask also the topography. Default True
            shift: Number of voxels shifted for the topology. Default 1.

        Returns:
              numpy.ndarray: masked regular grid
        """

        self.mask_matrix_pad = []
        series_type = self.stack.df['BottomRelation']
        for e, mask_series in enumerate(self.mask_matrix):
            mask_series_reshape = mask_series.reshape(self.grid.regular_grid.resolution)

            interfaces_from_block_bottoms = find_interfaces_from_block_bottoms(
                block=mask_series_reshape,
                value=True,
                shift=shift
            )
            mask_pad = mask_series_reshape + interfaces_from_block_bottoms

            if series_type[e] == 'Fault':
                mask_pad = np.invert(mask_pad)

            if mask_topography and self.grid.regular_grid.mask_topo.size != 0:
                mask_pad *= np.invert(self.grid.regular_grid.mask_topo)
            self.mask_matrix_pad.append(mask_pad)
        return self.mask_matrix_pad

    def compute_all_surfaces(self, **kwargs):
        """Compute all surfaces of the model given the geological features rules.

        Args:
            **kwargs: :any:`skimage.measure.marching_cubes` args (see below)

        Returns:
            list: vertices and edges

        See Also:
            :meth:`gempy.core.solution.Solution.compute_marching_cubes_regular_grid`

        """
        self.vertices = []
        self.edges = []
        mask_topography = kwargs.pop('mask_topography', True)
        masked_marching_cubes = kwargs.pop('masked_marching_cubes', True)
        self.mask_matrix_pad = self.padding_mask_matrix(
            mask_topography=mask_topography
        )
        series_type = self.stack.df['BottomRelation']
        series_number = 0
        active_indices = self.surfaces.df.groupby('isActive').groups[True]
        rescale = kwargs.pop('rescale', False)

        # We loop the scalar fields
        for e, scalar_field in enumerate(self.scalar_field_matrix):
            # Drop
            mask_array, sfas = self._prepare_marching_cubes_args(
                stack_number=e,
                masked_marching_cubes=masked_marching_cubes,
                series_type=series_type
            )
            
            for level in sfas:
                simpleces, vertices = self._try_compute_marching_cubes_on_the_regular_grid(
                    level=level,
                    mask_array=mask_array,
                    rescale=rescale,
                    s_n=series_number,
                    scalar_field=scalar_field,
                    kwargs=kwargs
                )
                series_number = self.set_vertices_edges(active_indices, simpleces, series_number, vertices)
        return self.vertices, self.edges

    def _try_compute_marching_cubes_on_the_regular_grid(self, level, mask_array, rescale, s_n, scalar_field, kwargs):
        try:
            v, s, norm, val = self.compute_marching_cubes_regular_grid(
                level, scalar_field, mask_array, rescale=rescale, **kwargs)
        except Exception as e:
            warnings.warn('Surfaces not computed due to: ' + str(
                e) + '. The surface is: Series: ' + str(e) +
                          '; Surface Number:' + str(s_n))
            v = np.nan
            s = np.nan
        return s, v

    def set_vertices_edges(self, active_indices, s, s_n, v):
        self.vertices.append(v)
        self.edges.append(s)
        idx = active_indices[s_n]
        self.surfaces.df.loc[idx, 'vertices'] = [v]
        self.surfaces.df.loc[idx, 'edges'] = [s]
        s_n += 1
        return s_n

    def _prepare_marching_cubes_args(self, stack_number, masked_marching_cubes, series_type):

        sfas = self.scalar_field_at_surface_points[stack_number]
        sfas = sfas[np.nonzero(sfas)]
        if masked_marching_cubes is True:
            if series_type[stack_number - 1] == 'Onlap' and series_type[stack_number - 2] == 'Erosion':
                mask_array = self.mask_matrix_pad[stack_number - 1]
            else:
                mask_array = self.mask_matrix_pad[stack_number]

        else:
            mask_array = None
        return mask_array, sfas




import warnings
try:
    import qgrid
except ImportError:
    warnings.warn('qgrid package is not installed. No interactive dataframes available.')
#
#
# # =============================================================
# # This are functions that modify the data objects independently
#
# def create_faults_qgrid(faults_object):
#     qgrid_widget = qgrid.show_grid(faults_object.df,
#                                    show_toolbar=False,
#                                    column_options={'editable': True},
#                                    column_definitions={'isFinite': {'editable': False}})
#
#     def handle_set_is_fault(event, widget, debug=False):
#         if debug is True:
#             print(event)
#             print(widget)
#
#         if event['column'] == 'isFault':
#             idx = event['index']
#             #      cat_idx = qgrid_widget.df.loc[idx, 'series_names']
#
#             faults_object.set_is_fault([idx])
#             qgrid_widget._update_df()
#
#     qgrid_widget.on('cell_edited', handle_set_is_fault)
#
#     return qgrid_widget
#
#
# def create_rescaling_data_qgrid(rescaling_object):
#     qgrid_widget = qgrid.show_grid(rescaling_object.df,
#                                    show_toolbar=False,
#                                    grid_options={'sortable': False, 'highlightSelectedCell': True})
#
#     def handle_row_edit(event, widget, debug=False):
#         if debug is True:
#             print(event)
#             print(widget)
#             print(qgrid_widget._df)
#
#         if event['column'] == 'centers':
#             try:
#                 value = event['new']
#                 assert value.shape[0] is 3
#
#                 #rescaling_object.df.loc['values', event['column']] = value
#                 rescaling_object.modify_rescaling_parameters(event['column'], value)
#                 qgrid_widget._update_df()
#
#             except AssertionError:
#                 print('centers length must be 3: XYZ')
#                 qgrid_widget._update_df()
#
#     qgrid_widget.on('cell_edited', handle_row_edit)
#     return qgrid_widget
#
#
# def create_options_qgrid(options_object):
#     qgrid_widget = qgrid.show_grid(options_object.df,
#                                    show_toolbar=False,
#                                    grid_options={'sortable': False, 'highlightSelectedCell': True})
#
#     def handle_row_edit(event, widget, debug=False):
#         if debug is True:
#             print(event)
#             print(widget)
#             print(qgrid_widget._df)
#
#         if event['column'] == 'verbosity':
#             import numpy as np
#             value = np.fromstring(event['new'][1:-1], sep=',')
#         else:
#             value = event['new']
#
#         options_object.df.loc['values', event['column']] = value
#         qgrid_widget._update_df()
#
#     qgrid_widget.on('cell_edited', handle_row_edit)
#     return qgrid_widget
#
#
# def create_kriging_parameters_qgrid(kriging_parameters_object):
#     qgrid_widget = qgrid.show_grid(kriging_parameters_object.df,
#                                    show_toolbar=False,
#                                    grid_options={'sortable': False, 'highlightSelectedCell': True},
#                                    )
#
#     def handle_row_edit(event, widget, debug=False):
#         import numpy as np
#
#         if debug is True:
#             print(event)
#             print(widget)
#             print(qgrid_widget._df)
#
#         if event['column'] == 'drift equations':
#             value = np.fromstring(event['new'][1:-1], sep=',')
#             try:
#                 assert value.shape[0] is kriging_parameters_object.structure.df.loc[
#                     'values', 'len series surface_points'].shape[0]
#
#                 kriging_parameters_object.df.loc['values', event['column']] = value
#                 qgrid_widget._update_df()
#
#             except AssertionError:
#                 print('u_grade length must be the same as the number of series')
#                 qgrid_widget._update_df()
#
#         else:
#
#             kriging_parameters_object.df.loc['values', event['column']] = event['new']
#             qgrid_widget._update_df()
#
#     qgrid_widget.on('cell_edited', handle_row_edit)
#     return qgrid_widget
#
#
# def create_faults_relations_qgrid(faults_object):
#     # We need to add the qgrid special columns to categories
#     faults_object.faults_relations_df.columns = faults_object.faults_relations_df.columns.add_categories(
#         ['index', 'qgrid_unfiltered_index'])
#
#     qgrid_widget = qgrid.show_grid(faults_object.faults_relations_df,
#                                    show_toolbar=False)
#
#     def handle_set_fault_relation(event, widget, debug=False):
#         if debug is True:
#             print(event)
#             print(widget)
#
#         faults_object.faults_relations_df = qgrid_widget.get_changed_df()
#         qgrid_widget._update_df()
#
#     qgrid_widget.on('cell_edited', handle_set_fault_relation)
#     return qgrid_widget
#
#
# def create_surface_points_qgrid(surface_points_object):
#     if surface_points_object.df.shape[0] == 0:
#         # TODO DEBUG: I am not sure that formations always has at least one entry. Check it
#         surface_points_object.add_surface_points(0, 0, 0, surface_points_object.formations.df['formation'].iloc[0])
#
#     qgrid_widget = qgrid.show_grid(
#         surface_points_object.df,
#         show_toolbar=True,
#         grid_options={'sortable': False, 'highlightSelectedCell': True},
#         column_options={'editable': False},
#         column_definitions={'X': {'editable': True},
#                             'Y': {'editable': True},
#                             'Z': {'editable': True},
#                             'surface': {'editable': True}})
#
#     def handle_row_surface_points_add(event, widget, debug=False):
#         if debug is True:
#             print(event)
#             print(widget)
#             print(qgrid_widget._df)
#         idx = event['index']
#         xyzs = qgrid_widget._df.loc[idx, ['X', 'Y', 'Z', 'surface']]
#
#         self._geo.add_surface_points(*xyzs)
#         qgrid_widget._update_df()
#
#     def handle_row_surface_points_delete(event, widget, debug=False):
#         if debug is True:
#             print(event)
#             print(widget)
#         idx = event['indices']
#
#         surface_points_object.del_surface_points(idx)
#         qgrid_widget._update_df()
#
#     def handle_cell_surface_points_edit(event, widget, debug=False):
#         if debug is True:
#             print(event)
#             print(widget)
#
#         column = event['column']
#         idx = event['index']
#         value = event['new']
#
#         surface_points_object.modify_surface_points(idx, **{column: value})
#
#         qgrid_widget._update_df()
#
#     qgrid_widget.on('row_removed', handle_row_surface_points_delete)
#     qgrid_widget.on('row_added', handle_row_surface_points_add)
#     qgrid_widget.on('cell_edited', handle_cell_surface_points_edit)
#     return qgrid_widget
#
#
# def create_orientations_qgrid(orientations_object):
#     if orientations_object.df.shape[0] == 0:
#         # TODO DEBUG: I am not sure that formations always has at least one entry. Check it
#         orientations_object.add_orientations(0, 0, 0,
#                                              orientations_object.formations.df['formation'].iloc[0],
#                                              0, 0, 1,
#                                              )
#
#     qgrid_widget = qgrid.show_grid(
#         orientations_object.df,
#         show_toolbar=True,
#         grid_options={'sortable': False, 'highlightSelectedCell': True},
#         column_options={'editable': False},
#         column_definitions={'X': {'editable': True},
#                             'Y': {'editable': True},
#                             'Z': {'editable': True},
#                             'G_x': {'editable': True},
#                             'G_y': {'editable': True},
#                             'G_z': {'editable': True},
#                             'azimuth': {'editable': True},
#                             'dip': {'editable': True},
#                             'pole': {'editable': True},
#                             'surface': {'editable': True}})
#
#     def handle_row_orientations_add(event, widget, debug=False):
#         if debug is True:
#             print(event)
#             print(widget)
#             print(qgrid_widget._df)
#         idx = event['index']
#         xyzs = qgrid_widget._df.loc[idx, ['X', 'Y', 'Z','surface', 'G_x', 'G_y', 'G_z']]
#
#         orientations_object.add_orientation(*xyzs)
#         qgrid_widget._update_df()
#
#     def handle_row_orientations_delete(event, widget, debug=False):
#         if debug is True:
#             print(event)
#             print(widget)
#         idx = event['indices']
#
#         orientations_object.del_orientation(idx)
#         qgrid_widget._update_df()
#
#     def handle_cell_orientations_edit(event, widget, debug=False):
#         if debug is True:
#             print(event)
#             print(widget)
#
#         column = event['column']
#         idx = event['index']
#         value = event['new']
#
#         orientations_object.modify_orientation(idx, **{column: value})
#         qgrid_widget._update_df()
#
#     qgrid_widget.on('row_removed', handle_row_orientations_delete)
#     qgrid_widget.on('row_added', handle_row_orientations_add)
#     qgrid_widget.on('cell_edited', handle_cell_orientations_edit)
#     return qgrid_widget
#
#
# def create_formations_qgrid(formation_object):
#     if formation_object.df.shape[0] == 0:
#         # TODO DEBUG: I am not sure that formations always has at least one entry. Check it
#         formation_object.set_surfaces_names(['surface1'])
#
#     qgrid_widget = qgrid.show_grid(formation_object.df, show_toolbar=True,
#                                    column_options={'editable': True},
#                                    column_definitions={'id': {'editable': False},
#                                                        'basement': {'editable': False}})
#
#     def handle_row_formation_add(event, widget, debug=False):
#         if debug is True:
#             print(event)
#             print(widget)
#         idx = event['index']
#         formation_object.add_surface(['surface' + str(idx)])
#         qgrid_widget._update_df()
#
#     def handle_row_formation_delete(event, widget, debug=False):
#         if debug is True:
#             print(event)
#             print(widget)
#         idx = event['indices']
#         formation_object.delete_surface(idx)
#         qgrid_widget._update_df()
#
#     def handle_cell_formation_edit(event, widget, debug=False):
#         if debug is True:
#             print(event)
#             print(widget)
#         if event['column'] == 'formation':
#             formation_object.rename_formations(event['old'], event['new'])
#         if event['column'] == 'series':
#             idx = event['index']
#             new_series = event['new']
#             formation_object.map_series({new_series:formation_object.df.loc[idx, ['formation']]})
#         if event['column'] == 'isBasement':
#             idx = event['index']
#             formation_object.set_basement(formation_object.df.loc[idx, ['formation']])
#         qgrid_widget._update_df()
#
#     qgrid_widget.on('row_removed', handle_row_formation_delete)
#     qgrid_widget.on('row_added', handle_row_formation_add)
#     qgrid_widget.on('cell_edited', handle_cell_formation_edit)
#
#     return qgrid_widget
#
#
# def create_series_qgrid(series_object):
#     # I need to do a serious hack because qgrid does not accept categorical index yet
#     qgrid_widget = qgrid.show_grid(
#         series_object.df.reset_index().rename(columns={'index': 'series_names'}).astype({'series_names': str}),
#         show_toolbar=True,
#         column_options={'editable': True},
#         column_definitions={'order_series': {'editable': False},
#                             })
#
#     def handle_row_series_add(event, widget, debug=False):
#         if debug is True:
#             print(event)
#             print(widget)
#             print(series_object.df.reset_index())
#
#         idx = event['index']
#         series_object.add_series(['series' + str(idx)])
#         qgrid_widget.df = series_object.df.reset_index().rename(columns={'index': 'series_names'}).astype(
#             {'series_names': str})
#         qgrid_widget._update_df()
#
#     def handle_row_series_delete(event, widget, debug=False):
#         if debug is True:
#             print(event)
#             print(widget)
#         idx = event['indices']
#         cat_idx = qgrid_widget.df.loc[idx, 'series_names']
#
#         series_object.delete_series(cat_idx)
#
#         qgrid_widget.df = series_object.df.reset_index().rename(columns={'index': 'series_names'}).astype(
#             {'series_names': str})
#         qgrid_widget._update_df()
#
#     def handle_cell_series_edit(event, widget, debug=False):
#         idx = event['index']
#         cat_idx = qgrid_widget.df.loc[idx, 'series_names']
#         if debug is True:
#             print(event)
#             print(widget)
#             print(cat_idx)
#             print(series_object.df.index)
#         if event['column'] == 'series_names':
#             series_object.rename_series({event['old']: event['new']})
#         if event['column'] == 'BottomRelation':
#             series_object.df.loc[cat_idx, 'BottomRelation'] = event['new']
#
#         qgrid_widget.df = series_object.df.reset_index().rename(columns={'index': 'series_names'}).astype(
#             {'series_names': str})
#         qgrid_widget._update_df()
#
#     qgrid_widget.on('row_removed', handle_row_series_delete)
#     qgrid_widget.on('row_added', handle_row_series_add)
#     qgrid_widget.on('cell_edited', handle_cell_series_edit)
#
#     return qgrid_widget



from typing import Union

import numpy as np

from gempy_legacy.core.grid_modules import grid_types, topography
from gempy_legacy.utils import docstring as ds
from gempy_legacy.utils.meta import _setdoc_pro, _setdoc


@_setdoc_pro([grid_types.RegularGrid.__doc__, grid_types.CustomGrid.__doc__])
class Grid(object):
    """ Class to generate grids.

    This class is used to create points to evaluate the geological model. 
    This class serves a container which transmits the XYZ coordinates to the
    interpolator. There are several type of grid objects feeding into the Grid class

    Args:
         **kwargs: See below

    Keyword Args:
         regular (:class:`gempy.core.grid_modules.grid_types.RegularGrid`): [s0]
         custom (:class:`gempy.core.grid_modules.grid_types.CustomGrid`): [s1]
         topography (:class:`gempy.core.grid_modules.grid_types.Topography`): [s2]
         sections (:class:`gempy.core.grid_modules.grid_types.Sections`): [s3]
         gravity (:class:`gempy.core.grid_modules.grid_types.Gravity`):

    Attributes:
        values (np.ndarray): coordinates where the model is going to be evaluated. This is the coordinates
         concatenation of all active grids.
        values_r (np.ndarray): rescaled coordinates where the model is going to be evaluated
        length (np.ndarray):I a array which contain the slicing index for each grid type in order. The first element will
         be 0, the second the length of the regular grid; the third custom and so on. This can be used to slice the
         solutions correspondent to each of the grids
        grid_types(np.ndarray[str]): names of the current grids of GemPy
        active_grids(np.ndarray[bool]): boolean array which controls which type of grid is going to be computed and
         hence on the property `values`.
        regular_grid (:class:`gempy.core.grid_modules.grid_types.RegularGrid`)
        custom_grid (:class:`gempy.core.grid_modules.grid_types.CustomGrid`)
        topography (:class:`gempy.core.grid_modules.grid_types.Topography`)
        sections (:class:`gempy.core.grid_modules.grid_types.Sections`)
        gravity_grid (:class:`gempy.core.grid_modules.grid_types.Gravity`)
    """

    def __init__(self, **kwargs):

        self.values = np.empty((0, 3))
        self.values_r = np.empty((0, 3))
        self.length = np.empty(0)
        self.grid_types = np.array(['regular', 'custom', 'topography', 'sections', 'centered'])
        self.active_grids = np.zeros(5, dtype=bool)
        # All grid types must have values

        # Init optional grids
        self.custom_grid = None
        self.custom_grid_grid_active = False
        self.topography = None
        self.topography_grid_active = False
        self.sections_grid_active = False
        self.centered_grid = None
        self.centered_grid_active = False

        # Init basic grid empty
        self.regular_grid = self.create_regular_grid(set_active=False, **kwargs)
        self.regular_grid_active = False

        # Init optional sections
        self.sections = grid_types.Sections(regular_grid=self.regular_grid)

        self.update_grid_values()

    def __str__(self):
        return 'Grid Object. Values: \n' + np.array2string(self.values)

    def __repr__(self):
        return 'Grid Object. Values: \n' + np.array_repr(self.values)

    @_setdoc(grid_types.RegularGrid.__doc__)
    def create_regular_grid(self, extent=None, resolution=None, set_active=True, *args, **kwargs):
        """
        Set a new regular grid and activate it.

        Args:
            extent (np.ndarray): [x_min, x_max, y_min, y_max, z_min, z_max]
            resolution (np.ndarray): [nx, ny, nz]

        RegularGrid Docs
        """
        self.regular_grid = grid_types.RegularGrid(extent, resolution, **kwargs)
        if set_active is True:
            self.set_active('regular')
        return self.regular_grid

    @_setdoc_pro(ds.coord)
    def create_custom_grid(self, custom_grid: np.ndarray):
        """
        Set a new regular grid and activate it.

        Args:
            custom_grid (np.array): [s0]

        """
        self.custom_grid = grid_types.CustomGrid(custom_grid)
        self.set_active('custom')

    def create_topography(self, source='random', **kwargs):
        """Create a topography grid and activate it.

        Args:
            source:
                * 'gdal':  Load topography from a raster file.
                * 'random': Generate random topography (based on a fractal grid).
                * 'saved': Load topography that was saved with the topography.save() function.
                  This is useful after loading and saving a heavy raster file with gdal once or after saving a
                  random topography with the save() function. This .npy file can then be set as topography.

        Keyword Args:
            source = 'gdal':
                * filepath:   path to raster file, e.g. '.tif', (for all file formats see
                  https://gdal.org/drivers/raster/index.html)

            source = 'random':
                * fd:         fractal dimension, defaults to 2.0
                * d_z:        maximum height difference. If none, last 20% of the model in z direction
                * extent:     extent in xy direction. If none, geo_model.grid.extent
                * resolution: desired resolution of the topography array. If none, geo_model.grid.resoution

            source = 'saved':
                * filepath:   path to the .npy file that was created using the topography.save() function

        Returns:
             :class:gempy.core.data.Topography
        """
        self.topography = topography.Topography(self.regular_grid)

        if source == 'random':
            self.topography.load_random_hills(**kwargs)
        elif source == 'gdal':
            filepath = kwargs.get('filepath', None)
            if filepath is not None:
                self.topography.load_from_gdal(filepath)
            else:
                print('to load a raster file, a path to the file must be provided')
        elif source == 'saved':
            filepath = kwargs.get('filepath', None)
            if filepath is not None:
                self.topography.load_from_saved(filepath)
            else:
                print('path to .npy file must be provided')
        elif source == 'numpy':
            array = kwargs.get('array', None)
            self.topography.set_values(array)
        else:
            raise AttributeError('source must be random, gdal or saved')

        self.set_active('topography')

    @_setdoc(grid_types.Sections.__doc__)
    def create_section_grid(self, section_dict):
        self.sections = grid_types.Sections(regular_grid=self.regular_grid, section_dict=section_dict)
        self.set_active('sections')
        return self.sections

    @_setdoc(grid_types.CenteredGrid.set_centered_grid.__doc__)
    def create_centered_grid(self, centers, radius, resolution=None):
        """Initialize gravity grid. Deactivate the rest of the grids"""
        self.centered_grid = grid_types.CenteredGrid(centers, radius, resolution)
        # self.active_grids = np.zeros(4, dtype=bool)
        self.set_active('centered')

    def deactivate_all_grids(self):
        """
        Deactivates the active grids array
        :return:
        """
        self.active_grids = np.zeros(5, dtype=bool)
        self.update_grid_values()
        return self.active_grids

    def set_active(self, grid_name: Union[str, np.ndarray]):
        """
        Set active a given or several grids
        Args:
            grid_name (str, list):

        """
        where = self.grid_types == grid_name
        self.active_grids[where] = True
        self.update_grid_values()
        return self.active_grids

    def set_inactive(self, grid_name: str):
        where = self.grid_types == grid_name
        self.active_grids *= ~where
        self.update_grid_values()
        return self.active_grids

    def update_grid_values(self):
        """
        Copy XYZ coordinates from each specific grid to Grid.values for those which are active.

        Returns:
            values

        """
        self.length = np.empty(0)
        self.values = np.empty((0, 3))
        lengths = [0]
        try:
            for e, grid_types in enumerate(
                    [self.regular_grid, self.custom_grid, self.topography, self.sections, self.centered_grid]):
                if self.active_grids[e]:
                    self.values = np.vstack((self.values, grid_types.values))
                    lengths.append(grid_types.values.shape[0])
                else:
                    lengths.append(0)
        except AttributeError:
            raise AttributeError('Grid type does not exist yet. Set the grid before activating it.')

        self.length = np.array(lengths).cumsum()
        return self.values

    def get_grid_args(self, grid_name: str):
        assert type(grid_name) is str, 'Only one grid type can be retrieved'
        assert grid_name in self.grid_types, 'possible grid types are ' + str(self.grid_types)
        where = np.where(self.grid_types == grid_name)[0][0]
        return self.length[where], self.length[where + 1]

    def get_grid(self, grid_name: str):
        assert type(grid_name) is str, 'Only one grid type can be retrieved'

        l_0, l_1 = self.get_grid_args(grid_name)
        return self.values[l_0:l_1]

    def get_section_args(self, section_name: str):
        # assert type(section_name) is str, 'Only one section type can be retrieved'
        l0, l1 = self.get_grid_args('sections')
        where = np.where(self.sections.names == section_name)[0][0]
        return l0 + self.sections.length[where], l0 + self.sections.length[where + 1]



""" setters API

"""
from gempy_legacy import get_data
from gempy_legacy.utils.meta import _setdoc, _setdoc_pro
from gempy_legacy.utils import docstring as ds
from gempy_legacy.core.model import Model, InterpolatorModel
from typing import Union
import warnings
import numpy as np

# This warning comes from numpy complaining about a aesara optimization
warnings.filterwarnings("ignore",
                        message='.* a non-tuple sequence for multidimensional '
                                'indexing is deprecated; use*.',
                        append=True)


@_setdoc([InterpolatorModel.__doc__])
@_setdoc_pro([Model.__doc__, ds.compile_aesara, ds.aesara_optimizer])
def set_interpolator(geo_model: Model, output: list = None, compile_aesara: bool = True,
                     aesara_optimizer=None, verbose: list = None, grid=None, type_=None,
                     update_structure=True, update_kriging=True,
                     **kwargs):
    """
    Method to create a graph and compile the aesara code to compute the interpolation.

    Args:
        geo_model (:class:`gempy.core.model.Project`): [s0]
        output (list[str:{geo, grav}]): type of interpolation.
        compile_aesara (bool): [s1]
        aesara_optimizer (str {'fast_run', 'fast_compile'}): [s2]
        verbose:
        update_kriging (bool): reset kriging values to its default.
        update_structure (bool): sync Structure instance before setting aesara graph.

    Keyword Args:
        -  pos_density (Optional[int]): Only necessary when type='grav'. Location on the Surfaces().df
         where density is located (starting on id being 0).
        - Vs
        - pos_magnetics

    Returns:

    """
    # output = list(output)
    if output is None:
        output = ['geology']

    if type(output) is not list:
        raise TypeError('Output must be a list.')

    # TODO Geology is necessary for everthing?
    if 'gravity' in output and 'geology' not in output:
        output.append('geology')

    if 'magnetics' in output and 'geology' not in output:
        output.append('geology')

    if type_ is not None:
        warnings.warn('type warn is going to be deprecated. Use output insted', FutureWarning)
        output = type_

    if aesara_optimizer is not None:
        geo_model._additional_data.options.df.at['values', 'aesara_optimizer'] = aesara_optimizer
    if verbose is not None:
        geo_model._additional_data.options.df.at['values', 'verbosity'] = verbose
    if 'dtype' in kwargs:
        geo_model._additional_data.options.df.at['values', 'dtype'] = kwargs['dtype']
    if 'device' in kwargs:
        geo_model._additional_data.options.df.at['values', 'device'] = kwargs['device']

    # TODO add kwargs
    geo_model._rescaling.rescale_data()
    geo_model.update_additional_data(update_structure=update_structure, update_kriging=update_kriging)
    geo_model.update_to_interpolator()
    geo_model._surface_points.sort_table()
    geo_model._orientations.sort_table()

    geo_model._interpolator.create_aesara_graph(geo_model._additional_data, inplace=True,
                                                output=output, **kwargs)

    if 'gravity' in output:
        pos_density = kwargs.get('pos_density', 1)
        tz = kwargs.get('tz', 'auto')
        if geo_model._grid.centered_grid is not None:
            geo_model._interpolator.set_aesara_shared_gravity(tz, pos_density)

    if 'magnetics' in output:
        pos_magnetics = kwargs.get('pos_magnetics', 1)
        Vs = kwargs.get('Vs', 'auto')
        incl = kwargs.get('incl')
        decl = kwargs.get('decl')
        B_ext = kwargs.get('B_ext', 52819.8506939139e-9)
        if geo_model._grid.centered_grid is not None:
            geo_model._interpolator.set_aesara_shared_magnetics(Vs, pos_magnetics, incl, decl, B_ext)

    if 'topology' in output:
        # This id is necessary for topology
        id_list = geo_model._surfaces.df.groupby('isFault').cumcount() + 1
        geo_model.add_surface_values(id_list, 'topology_id')
        geo_model._interpolator.set_aesara_shared_topology()

        # TODO it is missing to pass to aesara the position of topology_id

    if compile_aesara is True:
        geo_model._interpolator.set_all_shared_parameters(reset_ctrl=True)

        geo_model._interpolator.compile_th_fn_geo(inplace=True, grid=grid)
    else:
        if grid == 'shared':
            geo_model._interpolator.set_aesara_shared_grid(grid)

    print('Kriging values: \n', geo_model._additional_data.kriging_data)
    return geo_model._interpolator


@_setdoc_pro([Model.__doc__])
def set_geometric_data(geo_model: Model, surface_points_df=None,
                       orientations_df=None, **kwargs):
    """ Function to set directly pandas.Dataframes to the gempy geometric data objects

    Args:
        geo_model: [s0]
        surface_points_df:  A pn.Dataframe object with X, Y, Z, and surface columns
        orientations_df: A pn.Dataframe object with X, Y, Z, surface columns and pole or orientation columns
        **kwargs:

    Returns:
        Modified df
    """

    r_ = None

    if surface_points_df is not None:
        geo_model.set_surface_points(surface_points_df, **kwargs)
        r_ = 'surface_points'

    elif orientations_df is not None:
        geo_model.set_orientations(orientations_df, **kwargs)
        r_ = 'data' if r_ == 'surface_points' else 'orientations'

    else:
        raise AttributeError('You need to pass at least one dataframe')

    return get_data(geo_model, itype=r_)


def set_orientation_from_surface_points(geo_model, indices_array):
    """
    Create and set orientations from at least 3 points of the :attr:`gempy.data_management.InputData.surface_points`
     Dataframe

    Args:
        geo_model (:class:`Model`):
        indices_array (array-like): 1D or 2D array with the pandas indices of the
          :attr:`surface_points`. If 2D every row of the 2D matrix will be used to create an
          orientation


    Returns:
        :attr:`orientations`: Already updated inplace
    """

    if np.ndim(indices_array) == 1:
        indices = indices_array
        form = geo_model._surface_points.df['surface'].loc[indices].unique()
        assert form.shape[0] == 1, 'The interface points must belong to the same surface'
        form = form[0]

        ori_parameters = geo_model._orientations.create_orientation_from_surface_points(geo_model._surface_points, indices)
        geo_model.add_orientations(
            X=ori_parameters[0],
            Y=ori_parameters[1],
            Z=ori_parameters[2],
            orientation=ori_parameters[3:6],
            pole_vector=ori_parameters[6:9],
            surface=form
        )
    elif np.ndim(indices_array) == 2:
        for indices in indices_array:
            form = geo_model._surface_points.df['surface'].loc[indices].unique()
            assert form.shape[0] == 1, 'The interface points must belong to the same surface'
            form = form[0]
            ori_parameters = geo_model._orientations.create_orientation_from_surface_points(
                geo_model._surface_points, indices)
            geo_model.add_orientations(X=ori_parameters[0], Y=ori_parameters[1], Z=ori_parameters[2],
                                       orientation=ori_parameters[3:6], pole_vector=ori_parameters[6:9],
                                       surface=form)

    return geo_model._orientations


def select_nearest_surfaces_points(geo_model, surface_points, searchcrit):
    """
    Find the neighbour points of the same surface
    by given radius (radius-search) or fix number (knn).
    Parameters
    ----------
    geo_model : geo_model
        GemPy-model.
    surface_points: Pandas-dataframe
        Contains the dataframe of the (point-)data from the GemPy-model.
    searchcrit : int or float
        if is int: uses knn-search.
        if is float: uses radius-search.
    """

    from sklearn.neighbors import NearestNeighbors

    # extract surface names
    surfaces = np.unique(surface_points['surface'])
    neighbours = []
    # for each surface
    if isinstance(searchcrit, int):  # in case knn-search
        searchcrit = searchcrit + 1  # because the point itself is also found
        for s in range(surfaces.size):
            # extract point-ids
            i_surfaces = surface_points['surface'] == surfaces[s]
            # extract point coordinates
            p_surfaces = surface_points[i_surfaces][['X', 'Y', 'Z']]
            # create search-tree
            Tree = NearestNeighbors(n_neighbors=searchcrit)
            # add data to tree
            Tree.fit(p_surfaces)
            # find neighbours
            neighbours_surfaces = Tree.kneighbors(p_surfaces, n_neighbors=searchcrit,
                                                  return_distance=False)
            # add neighbours with initial index to total list
            for n in neighbours_surfaces:
                neighbours.append(p_surfaces.index[n])
    else:  # in case radius-search
        for s in range(surfaces.size):
            # extract point-ids
            i_surfaces = surface_points['surface'] == surfaces[s]
            # extract point coordinates
            p_surfaces = surface_points[i_surfaces][['X', 'Y', 'Z']]
            # create search-tree
            Tree = NearestNeighbors(radius=searchcrit)
            # add data to tree
            Tree.fit(p_surfaces)
            # find neighbours (attention: relativ index!)
            neighbours_surfaces = Tree.radius_neighbors(p_surfaces,
                                                        radius=searchcrit,
                                                        return_distance=False)
            # add neighbours with initial index to total list
            for n in neighbours_surfaces:
                neighbours.append(p_surfaces.index[n])
    return neighbours


def set_orientation_from_neighbours(geo_model, neighbours):
    """
    Calculates the orientation of one point with its neighbour points
    of the same surface.
    Parameters
    ----------
    geo_model : geo_model
        GemPy-model.
    neighbours : Int64Index
        point-neighbours-id, first id is the point itself.
    """

    from sklearn.preprocessing import normalize

    # compute normal vector for the point
    if neighbours.size > 2:
        # extract point coordinates
        coo = geo_model._surface_points.df.loc[neighbours][['X', 'Y', 'Z']]
        # calculates covariance matrix
        cov = np.cov(coo.T)
        # calculate normalized normal vector
        normvec = normalize(np.cross(cov[0].T, cov[1].T).reshape(1, -1))[0]
        # check orientation of normal vector (has to be oriented to sky)
        if normvec[2] < 0:
            normvec = normvec * (-1)
        # append to the GemPy-model
        geo_model.add_orientations(geo_model._surface_points.df['X'][neighbours[0]],
                                   geo_model._surface_points.df['Y'][neighbours[0]],
                                   geo_model._surface_points.df['Z'][neighbours[0]],
                                   geo_model._surface_points.df['surface'][neighbours[0]],
                                   normvec.tolist())
    # if computation is impossible set normal vector to default orientation
    else:
        print("orientation calculation of point" + str(neighbours[0]) + "is impossible")
        print("-> default vector is set [0,0,1]")
        geo_model.add_orientations(geo_model._surface_points.df['X'][neighbours[0]],
                                   geo_model._surface_points.df['Y'][neighbours[0]],
                                   geo_model._surface_points.df['Z'][neighbours[0]],
                                   geo_model._surface_points.df['surface'][neighbours[0]],
                                   orientation=[0, 0, 1])
    return geo_model._orientations


def set_orientation_from_neighbours_all(geo_model, neighbours):
    """
    Calculates the orientations for all points with given neighbours.
    Parameters
    ----------
    geo_model : geo_model
        GemPy-model.
    neighbours : list of Int64Index
        point-neighbours-IDs, the first id is the id of the point
        for which the orientation is calculated.
    """

    # compute normal vector for the points
    for n in neighbours:
        set_orientation_from_neighbours(geo_model, n)

    return geo_model._orientations



# These two lines are necessary only if GemPy is not installed
# Importing GemPy
import gempy_legacy as gp

# Importing auxiliary libraries
import numpy as np
import pytest
import os
import matplotlib.pyplot as plt

mm = gp.ImplicitCoKriging()
mm.add_surfaces(['surface1', 'foo1', 'foo2', 'foo3'])


def test_add_surface_points_raise_non_surface():
    with pytest.raises(TypeError, match=r'.*surface passed does not exist in the pandas categories..*'):
        mm.add_surface_points(400, 300, -500, 'surface5')


def test_add_surface():
    mm.add_surfaces(['a_foo1', 'a_foo2'])
    with pytest.raises(ValueError, match=r'.* not include old categories.*'):
        mm.add_surfaces('a_foo1')

    mm.add_surfaces('a_foo3')


def test_delete_surface():
    mm.delete_surfaces('foo1')
    mm.delete_surfaces(['surface1', 'foo2'])


def test_rename_surface():
    mm = gp.ImplicitCoKriging()
    mm.add_surfaces(['surface1', 'foo1', 'foo2', 'foo3'])
    mm.rename_surfaces({'foo1': 'changed'})
    assert mm._surfaces.df.loc[1, 'surface'] == 'changed'


def test_modify_order_surfaces():
    mm = gp.ImplicitCoKriging()
    mm.add_surfaces(['surface1', 'foo1', 'foo2', 'foo3'])
    mm.modify_order_surfaces(3, 2)
    assert mm._surfaces.df.iloc[2, 0] == 'foo2'


def test_add_surface_values():
    pass


def test_delete_surface_values():
    pass


def test_modify_surface_values():
    mm = gp.ImplicitCoKriging()
    mm.add_surfaces(['surface1', 'foo1', 'foo2', 'foo3'])
    mm.add_surface_points(400, 300, -500, 'foo2')
    print(mm._surface_points)
    mm.modify_surface_points(0, Y=800)
    print(mm._surface_points)


def test_set_surface_values():
    pass


def test_add_surface_points():
    mm = gp.ImplicitCoKriging()
    mm.add_surfaces(['surface1', 'foo1', 'foo2', 'foo3'])
    mm.add_surface_points(400, 300, -500, 'foo2')


def test_add_default_orientation():
    mm = gp.ImplicitCoKriging()
    mm.set_default_surfaces()
    mm.set_default_orientation()


def test_set_is_fault():
    mm = gp.ImplicitCoKriging()
    mm.add_features(['foo1', 'foo2', 'foo3'])
    assert (mm._faults.df.index == np.array(['Default series', 'foo1', 'foo2', 'foo3'])).all()
    assert (mm._faults.faults_relations_df.index == ['Default series', 'foo1', 'foo2', 'foo3']).all()
    mm.set_is_fault(['foo2'])
    assert mm._faults.faults_relations_df.loc['foo2', 'foo3'] == True
    assert mm._faults.faults_relations_df.iloc[2, 3] == True
    mm.set_is_fault(['foo2'], toggle=True)


def test_read_data():
    data_path = os.path.dirname(__file__)+'/../../examples/'
    model = gp.Model()
    model.read_data(path_i=data_path + "/data/input_data/tut_chapter1/simple_fault_model_points.csv",
                    path_o=data_path + "/data/input_data/tut_chapter1/simple_fault_model_orientations.csv")

    assert model._surface_points.df.shape[0] == 57

def test_add_surface_points_to_model():
    geo_model = gp.create_model('TestModel1')
    gp.init_data(geo_model, extent=[0, 800, 0, 200, -600, 0], resolution=[100, 100, 100])
    geo_model.set_default_surfaces()

    geo_model.add_surface_points(X=223, Y=0.01, Z=-94, surface='surface1')
    geo_model.add_surface_points(X=458, Y=0, Z=-107, surface='surface1')
    geo_model.add_surface_points(X=612, Y=0, Z=-14, surface='surface1')
    geo_model.add_orientations(X=350, Y=0, Z=-300, surface='surface1', pole_vector=(0, 0, 1))

    geo_model.add_surface_points(X=225, Y=1, Z=-269, surface='surface2')
    geo_model.add_surface_points(X=459, Y=1, Z=-279, surface='surface2')

    #gp.plot_2d(geo_model, cell_number=5, legend='force')
    #plt.show()



class MetaData(object):
    """Class containing metadata of the project.

    Set of attributes and methods that are not related directly to the geological model but more with the project

    Args:
        project_name (str): Name of the project. This is used as the default value for some I/O actions

    Attributes:
        date (str): Time of the creation of the project
        project_name (str): Name of the project. This is used as the default value for some I/O actions

    """

    def __init__(self, project_name='default_project'):
        import datetime
        now = datetime.datetime.now()
        self.date = now.strftime(" %Y-%m-%d %H:%M")

        if project_name == 'default_project':
            project_name += self.date

        self.project_name = project_name



from typing import Union

from gempy_legacy.core.data_modules.stack import Stack
from gempy_legacy import Surfaces, Grid

try:
    import xarray as xr
    import subsurface
    from subsurface.structs.base_structures.common_data_utils import to_netcdf
except:
    print("Not subsurface compatibility available")



import numpy as np
import pandas as pd


class XSolution(object):
    """This class stores the output of the interpolation and the necessary objects
    to visualize and manipulate this data using xarray as backend.

    Depending on the activated grid (see :class:`Grid`) a different number of
     properties are returned returned:

    Args:
        grid (Grid): [s0]
        surfaces (Surfaces): [s1]
        series (Series): [s2]

    Attributes:
        grid (Grid)
        surfaces (Surfaces)
        series (Series)
        weights_vector (numpy.array): [s3]
        scalar_field_at_surface_points (numpy.array): [s4]
        block_at_surface_points (numpy.array): [s5]
        mask_at_surface_points (numpy.array): [s6]
        values_at_surface_points (numpy.array): [s7]
        lith_block (numpy.array): [s8]
        scalar_field_matrix (numpy.array): [s9]
        block_matrix (numpy.array): [s10]
        mask_matrix (numpy.array): [s11]
        mask_matrix_pad (numpy.array): mask matrix padded 2 block in order to guarantee that the layers intersect each
         other after marching cubes
        values_matrix (numpy.array): [s12]
        vertices (list[numpy.array]): [s13]
        edges (list[numpy.array]): [s14]
        geological_map (numpy.array): [s15]

    """

    def __init__(self, grid: Grid,
                 surfaces: Surfaces = None,
                 stack: Stack = None,
                 ):

        self.grid = grid
        self.stack = stack
        self.surfaces = surfaces  # Used to store ver/sim there

        # Define xarrays
        self.weights_vector = None
        self.at_surface_points = None
        self.s_regular_grid = None  # xr.Dataset()
        self.s_custom_grid = None  # xr.Dataset()
        self.s_topography = None  # xr.Dataset()
        self.s_at_surface_points = None  # xr.Dataset()
        self.s_sections = dict()
        self.meshes = None  # xr.Dataset()

    def set_meshes(self, surfaces: Surfaces = None):
        """Create xarray from the Surfaces object. In GemPy Engine we will set
         them directly.
        """

        surf_ver_sim = surfaces.df[['id', 'vertices', 'edges']]
        vertex = []
        simplex = []
        ids = []
        last_idx = 0
        self.extract_each_surface_representations(ids, last_idx, simplex,
                                                  surf_ver_sim, vertex)

        vertex_array = np.concatenate(vertex)
        simplex_array = np.concatenate(simplex)
        ids_array = np.concatenate(ids)

        self.meshes = subsurface.UnstructuredData.from_array(
            vertex=vertex_array,
            cells=simplex_array,
            attributes=pd.DataFrame(ids_array, columns=['id'])
        )

        return self.meshes

    @staticmethod
    def extract_each_surface_representations(ids, last_idx, simplex,
                                             surf_ver_sim, vertex):
        for index, row in surf_ver_sim.iterrows():
            v_ = row['vertices']
            e_ = row['edges'] + last_idx
            if v_ is not np.nan and e_ is not np.nan:
                i_ = np.ones(e_.shape[0]) * row['id']
                vertex.append(v_)
                simplex.append(e_)
                ids.append(i_)
                last_idx = e_[-20:].max() + 1

    def set_values(self,
                   values: list,
                   active_features=None,
                   surf_properties=None,
                   ):
        """ At this stage we should split values into the different grids

        Args:
            values:

        Returns:

        """

        # Get an array with all the indices for each grid
        l = self.grid.length

        coords_base, xyz = self.prepare_common_args(active_features,
                                                    surf_properties)
        self.weights_vector = values[3]

        if self.grid.active_grids[0]:
            self.set_values_to_regular_grid_(values, l[0], l[1], coords_base.copy())
        if self.grid.active_grids[1]:
            self.set_values_to_custom_grid(values, l[1], l[2], coords_base.copy(),
                                           xyz=xyz)
        if self.grid.active_grids[2]:
            self.set_values_to_topography(values, l[2], l[3], coords_base.copy())
        if self.grid.active_grids[3]:
            self.set_values_to_sections(values, l[3], l[4], coords_base.copy())
        if self.grid.active_grids[4]:
            self.set_values_to_centered()

        # TODO: Add xyz from surface points
        self.set_values_to_surface_points_(values, l[-1], coords_base, xyz=None)

    def prepare_common_args(self, active_features, surf_properties):
        if active_features is None and self.stack is not None:
            active_features = self.stack.df.groupby('isActive').get_group(True).index
        if surf_properties is None and self.surfaces is not None:
            surf_properties = self.surfaces.properties_val
        coords_base = dict()
        if active_features is not None:
            coords_base['Features'] = active_features.to_list()
        if surf_properties is not None:
            coords_base['Properties'] = surf_properties.to_list()
        if self.grid.custom_grid is not None:
            xyz = self.grid.custom_grid.values
        else:
            xyz = None
        return coords_base, xyz

    def set_values_to_centered(self):
        return

    def set_values_to_surface_points_(self, values, l0, coords_base, xyz=None):
        coords = coords_base
        l1 = values[0].shape[-1]
        arrays = self.create_unstruct_xarray(values, l0, l1, xyz)

        self.s_at_surface_points = xr.Dataset(
            data_vars=arrays,
            coords=coords
        )
        return self.s_at_surface_points

    @staticmethod
    def create_struc_xarrays(values, l0, l1, res: Union[list, np.ndarray]):
        arrays = dict()

        n_dim = len(res)
        xyz = ['X', 'Y', 'Z'][:n_dim]
        if values[0] is not None:
            # This encompass lith_block and values matrix
            property_matrix = xr.DataArray(
                data=values[0][:, l0:l1].reshape(-1, *res),
                dims=['Properties', *xyz],
            )
            arrays['property_matrix'] = property_matrix

        if values[1] is not None:
            # This is the block matrix
            i, j, _ = values[1].shape
            block_matrix = xr.DataArray(
                data=values[1][:, :, l0:l1].reshape(i, j, *res),
                dims=['Features', 'Properties', *xyz],
            )
            arrays['block_matrix'] = block_matrix

            # Fault block?

        if values[4] is not None:
            # Scalar field matrix
            scalar_matrix = xr.DataArray(
                data=values[4][:, l0:l1].reshape(-1, *res),
                dims=['Features', *xyz],
            )
            arrays['scalar_field_matrix'] = scalar_matrix

        if values[6] is not None:
            # Mask matrix
            mask_matrix = xr.DataArray(
                data=values[6][:, l0:l1].reshape(-1, *res),
                dims=['Features', *xyz],
            )
            arrays['mask_matrix'] = mask_matrix

        if values[7] is not None:
            # Fault mask matrix
            fault_mask = xr.DataArray(
                data=values[7][:, l0:l1].reshape(-1, *res),
                dims=['Features', *xyz],
            )
            arrays['fault_mask'] = fault_mask

        return arrays

    @staticmethod
    def create_unstruct_xarray(values, l0, l1, xyz):
        arrays = dict()
        if xyz is not None:
            cartesian_matrix = xr.DataArray(
                data=xyz,
                dims=['Point', 'XYZ'],
                coords={'XYZ': ['X', 'Y', 'Z']}
            )
            arrays['cartesian_matrix'] = cartesian_matrix

        if values[0] is not None:
            # Values and lith block
            property_v3 = xr.DataArray(
                data=values[0][:, l0:l1],
                dims=['cell_attr', 'cell'],
            )

            arrays['property'] = property_v3

        if values[1] is not None:
            # block
            block_v3 = xr.DataArray(
                data=values[1][:, :, l0:l1],
                dims=['Features', 'cell_attr', 'cell'],
            )

            arrays['block'] = block_v3

        if values[4] is not None:
            # Scalar field
            scalar_field_v3 = xr.DataArray(
                data=values[4][:, l0:l1],
                dims=['Features', 'cell'],
            )
            arrays['scalar_field'] = scalar_field_v3

        if values[6] is not None:
            # Scalar field
            mask_v3 = xr.DataArray(
                data=values[6][:, l0:l1],
                dims=['Features', 'cell'],
            )
            arrays['mask'] = mask_v3

        return arrays

    def set_values_to_custom_grid(self, values: list, l0, l1,
                                  coords_base: dict, xyz=None):

        coords = coords_base
        arrays = self.create_unstruct_xarray(values, l0, l1, xyz=None)

        self.s_custom_grid = subsurface.UnstructuredData.from_array(
            vertex=xyz,
            cells="points",
            cells_attr=arrays,
            coords=coords,
            default_cells_attr_name="block"
        )

        return self.s_custom_grid

    def set_values_to_regular_grid_(self, values: list, l0, l1,
                                    coords_base: dict):

        coords = self.add_cartesian_coords(coords_base)

        arrays = self.create_struc_xarrays(values, l0, l1,
                                           self.grid.regular_grid.resolution)

        self.s_regular_grid = subsurface.StructuredData.from_dict(data_dict=arrays, coords=coords)

    def add_cartesian_coords(self, coords_base):
        coords = coords_base
        coords['X'] = self.grid.regular_grid.x
        coords['Y'] = self.grid.regular_grid.y
        coords['Z'] = self.grid.regular_grid.z
        return coords

    def set_values_to_topography(self,
                                 values: list,
                                 l0, l1,
                                 coords_base):
        coords = coords_base
        coords['X'] = self.grid.topography.x
        coords['Y'] = self.grid.topography.y
        resolution = self.grid.topography.resolution
        arrays = self.create_struc_xarrays(values, l0, l1, resolution)

        self.s_topography = subsurface.StructuredData.from_dict(data_dict=arrays, coords=coords)
        return self.s_topography

    def set_values_to_sections(self,
                               values: list,
                               l0, l1,
                               coords_base):
        coords = coords_base
        sections = self.grid.sections

        for e, axis_coord in enumerate(sections.generate_axis_coord()):
            resolution = sections.resolution[e]
            l0_s = sections.length[e]
            l1_s = sections.length[e + 1]
            name, xy = axis_coord
            coords['X'] = xy[:, 0]
            coords['Y'] = xy[:, 1]

            arrays = self.create_struc_xarrays(values, l0 + l0_s, l0 + l1_s, resolution)

            self.s_sections[name] = subsurface.StructuredData.from_dict(data_dict=arrays, coords=coords)
        return self.s_sections

    @property
    def data_structures(self):
        # TODO: Add sections
        args = [self.s_regular_grid, self.s_custom_grid, self.s_topography, self.meshes]
        names = ['regular_grid', 'custom_grid', 'topography', 'meshes']
        return zip(args, names)

    def to_netcdf(self, path, name, **kwargs):
        for a, n in self.data_structures:
            if a is not None:
                to_netcdf(a, f'{path}/{name}_{n}.nc', **kwargs)



import os
import sys

import gempy_legacy as gp
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt

from gempy_legacy.core.grid_modules.create_topography import GDAL_IMPORT
from gempy_legacy.core.grid_modules.topography import Topography
import pytest

import numpy as np

sys.path.append("../../..")
os.environ["aesara_FLAGS"] = "mode=FAST_RUN,device=cpu"
data_path = os.path.dirname(__file__) + '/../../input_data'

pytest.importorskip("osgeo.gdal")


@pytest.fixture(scope='module')
def artificial_grid(one_fault_model_no_interp):
    geo_model = one_fault_model_no_interp
    topo = Topography(geo_model._grid.regular_grid)
    topo.load_random_hills()
    print(topo.values, topo.values_2d)
    return topo


@pytest.mark.skipif(("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true") or
                    ("GITHUBACTION" in os.environ and os.environ["GITHUBACTION"] == "true"),
                    reason="Skipping this test on Travis CI because gdal.")
def test_real_grid_ales():
    resolution = [30, 20, 50]
    extent = np.array([729550.0, 751500.0, 1913500.0, 1923650.0, -50, 800.0])
    path_interf = data_path + "/2018_interf.csv"
    path_orient = data_path + "/2018_orient_clust_n_init5_0.csv"
    path_dem = data_path + "/_cropped_DEM_coarse.tif"

    geo_model = gp.create_model('Alesmodel')
    gp.init_data(geo_model, extent=extent, resolution=resolution,
                 path_i=path_interf,
                 path_o=path_orient)

    geo_model.set_topography(source='gdal', filepath=path_dem)

    p2d = gp.plot_2d(geo_model, section_names=['topography'], show_topography=True,
                     kwargs_topography={'hillshade': False, 'fill_contour': False})
    plt.show()

    p2d = gp.plot_2d(geo_model, section_names=['topography'], show_topography=True,
                     kwargs_topography={'hillshade': False, 'fill_contour': True})
    plt.show()

    p2d = gp.plot_2d(geo_model, section_names=['topography'], show_topography=True,
                     kwargs_topography={'hillshade': True, 'fill_contour': True})
    plt.show()

    if False:
        gp.map_stack_to_surfaces(geo_model, {'fault_left'     : ('fault_left'),
                                             'fault_right'    : ('fault_right'),
                                             'fault_lr'       : ('fault_lr'),
                                             'Trias_Series'   : ('TRIAS', 'LIAS'),
                                             'Carbon_Series'  : ('CARBO'),
                                             'Basement_Series': ('basement')}, remove_unused_series=True)

        geo_model.set_is_fault(['fault_right', 'fault_left', 'fault_lr'], change_color=True)
        gp.set_interpolator(geo_model,
                            output=['geology'], compile_aesara=True,
                            aesara_optimizer='fast_run', dtype='float64',
                            verbose=[])

        gp.compute_model(geo_model, compute_mesh=True)

        geo_model._grid.regular_grid.set_topography_mask(geo_model._grid.topography)

    gpv = gp.plot.plot_3d(geo_model,
                          plotter_type='basic', off_screen=True,
                          show_topography=True,
                          show_scalar=False,
                          show_lith=True,
                          show_surfaces=False,
                          kwargs_plot_structured_grid={'opacity'   : 1,
                                                       'show_edges': False},
                          ve=10,
                          image=True,
                          kwargs_plot_topography={'scalars': 'topography'})
    #
    # gpv.p.set_scale(zscale=10)
    #
    # img = gpv.p.show(screenshot=True)
    # plt.imshow(img[1])
    # plt.show()


def test_plot_2d_topography(one_fault_model_no_interp, artificial_grid):
    geo_model = one_fault_model_no_interp
    # geo_model._grid.topography = artificial_grid
    geo_model.set_topography()
    p2d = gp.plot_2d(geo_model, section_names=['topography'], show_topography=True,
                     kwargs_topography={'hillshade': False})
    plt.show()

    p2d = gp.plot_2d(geo_model, section_names=['topography'], show_topography=True,
                     kwargs_topography={'hillshade': True, 'fill_contour': False})
    plt.show()

    p2d = gp.plot_2d(geo_model, section_names=['topography'], show_topography=True,
                     kwargs_topography={'hillshade': True})
    plt.show()


def test_plot_3d_structure_topo2(unconformity_model_topo, artificial_grid):
    geo_model = unconformity_model_topo
    with pytest.raises(AssertionError):
        geo_model._grid.regular_grid.set_topography_mask(artificial_grid)

    # geo_model._grid.regular_grid.set_topography_mask(geo_model._grid.topography)

    p2d = gp.plot_2d(geo_model, section_names=['topography'], show_topography=True,
                     show_lith=False,
                     kwargs_topography={'hillshade': True})
    plt.show()

    gpv = gp.plot.plot_3d(unconformity_model_topo,
                          plotter_type='basic', off_screen=True,
                          show_topography=True,
                          show_scalar=False,
                          show_lith=True,
                          show_surfaces=True,
                          kwargs_plot_structured_grid={'opacity'   : .5,
                                                       'show_edges': True},
                          image=True,
                          kwargs_plot_topography={'scalars': 'topography'})

    # img = gpv.p.show(screenshot=True)
    # plt.imshow(img[1])
    # plt.show()



"""
Alesmodel: Plotting sections and maps.
======================================

"""

import gempy_legacy as gp
import numpy as np
import matplotlib.pyplot as plt
import os

# %%
cwd = os.getcwd()
if 'examples' not in cwd:
    data_path = os.getcwd() + '/examples'
else:
    data_path = cwd + '/../..'

path_interf = data_path + "/data/input_data/AlesModel/2018_interf.csv"
path_orient = data_path + "/data/input_data/AlesModel/2018_orient_clust_n_init5_0.csv"
path_dem = data_path + "/data/input_data/AlesModel/_cropped_DEM_coarse.tif"

# %% 
resolution = [100, 100, 100]
extent = np.array([729550.0, 751500.0, 1913500.0, 1923650.0, -1800.0, 800.0])
geo_model = gp.create_model('Alesmodel')
gp.init_data(geo_model, extent=extent, resolution=resolution,
             path_i=path_interf,
             path_o=path_orient)

# %% 
sdict = {'section1': ([732000, 1916000], [745000, 1916000], [200, 150])}
geo_model.set_section_grid(sdict)

# %% 
# sorting of lithologies
gp.map_stack_to_surfaces(geo_model, {'fault_left': ('fault_left'),
                                     'fault_right': ('fault_right'),
                                     'fault_lr': ('fault_lr'),
                                     'Trias_Series': ('TRIAS', 'LIAS'),
                                     'Carbon_Series': ('CARBO'),
                                     'Basement_Series': ('basement')}, remove_unused_series=True)

# %% 
colordict = {'LIAS': '#015482', 'TRIAS': '#9f0052', 'CARBO': '#ffbe00', 'basement': '#728f02',
             'fault_left': '#2a2a2a', 'fault_right': '#545454', 'fault_lr': '#a5a391'}
geo_model.surfaces.colors.change_colors(colordict)

# %% 
a = gp.plot_2d(geo_model, direction='y')

# %% 
geo_model.rescaling

# %% 
gp.plot.plot_section_traces(geo_model)

# %%
# Faults
# ''''''
# 

# %% 
geo_model.set_is_fault(['fault_right', 'fault_left', 'fault_lr'], change_color=True)

# %% 
gp.set_interpolator(geo_model,
                    output=['geology'], compile_aesara=True,
                    aesara_optimizer='fast_run', dtype='float64',
                    verbose=[])

# %%
# Topography
# ~~~~~~~~~~
# 

# %% 
geo_model.set_topography(source='gdal', filepath=path_dem)

# %% 
geo_model.surfaces

# %% 
_ = gp.compute_model(geo_model, compute_mesh=True, compute_mesh_options={'rescale': False})

# %% 
gp.plot_2d(geo_model, cell_number=[4], direction=['y'], show_topography=True,
           show_data=True)

# %% 
gp.plot_2d(geo_model, section_names=['topography'], show_data=False,
           show_boundaries=False)

# %%
# sphinx_gallery_thumbnail_number = 5
gp.plot_3d(geo_model)

# %%
# np.save('Ales_vert3', geo_model.solutions.vertices)
# np.save('Ales_edges3', geo_model.solutions.edges)

# %% 
# gp.plot.plot_ar(geo_model)

gp.save_model(geo_model)